/*!
\file  bkheap.c
\brief This file implements a min-bounded k-select heap data structure.
*/
#include "bkheap.h"




/** LOCAL FUNCTIONS **/
uint loc_h1_ (const idx_t id);
uint loc_h2_ (const idx_t id);
void loc_compact_overfow_ (bkheap * const heap);
void loc_reset_ (bkheap * const heap);
void loc_set_ (bkheap * const heap, const idx_t id);
int loc_isset_ (bkheap * const heap, const idx_t id);
void loc_unset_ (bkheap * const heap, const idx_t id);
void loc_unset_shallow_ (bkheap * const heap, const idx_t id);
void loc_reset_ (bkheap * const heap);

/**
 * Primary hash for the locator array
 */
inline uint
loc_h1_ (const idx_t id)
{
  return (id + 1) & HTMASK;
}

/**
 *  Secondary hash for the locator array
 */
inline uint
loc_h2_ (const idx_t id)
{
  uint h = id;
  // Secondary hashing discovered by Doug Lea
  h += (h << 9);
  h ^= (h >> 14);
  h += (h << 4);
  h ^= (h >> 10);

  return h & HTMASK;
}

/**
 * Compact locator hash overflow (remove deleted items to make room for new items)
 *
 */
inline void
loc_compact_overfow_ (bkheap * const heap)
{
  size_t i, j;
  idx_t id, key;
  uint h;

#ifdef HTHASHD
  heap->hz = 0;
  for (i = 0; i < heap->sz; ++i)
    {
      id = heap->data[i].key;
      h = loc_h1_ (id);

      if (id != (key = heap->loc[h]))
	{
	  if (BKHEAP_SKEY <= key)
	    {
	      heap->loc[h] = id;
	      continue;
	    }
	  h = loc_h2_ (id);
	  if (id != (key = heap->loc[h]))
	    {
	      if (BKHEAP_SKEY <= key)
		{
		  heap->loc[h] = id;
		  continue;
		}
	      heap->loc[HTSIZE + heap->hz] = id;
	      heap->hz++;
	    }
	}
    }
#else
  for (j = HTSIZE, i = HTSIZE; i < HTSIZE + heap->hz; ++i)
    {
      id = heap->data[i].key;
      h = loc_h1_ (id);
      if (id != (key = heap->loc[h]))
	{
	  if (BKHEAP_SKEY <= key)
	    {
	      heap->loc[h] = id;
	      continue;
	    }
	  else
	    {
	      heap->loc[j++] = id;
	    }
	}
    }
#endif
}


/**
 * Set an item in the locator hash
 */
inline void
loc_set_ (bkheap * const heap, const idx_t id)
{
  uint h;

  BKHEAP_LOOKUP (heap);

  h = loc_h1_ (id);
  if (BKHEAP_SKEY <= heap->loc[h] || id == heap->loc[h])
    {
      heap->loc[h] = id;
    }
  else
    {
#ifdef HTHASHD
      h = loc_h2_ (id);
      if (BKHEAP_SKEY <= heap->loc[h] || id == heap->loc[h])
	{
	  heap->loc[h] = id;
	}
      else
	{
	  BKHEAP_COLLISION (heap);
	    /** add to locator overflow */
	  if (heap->hz == heap->maxsz)
	    {
	      loc_compact_overfow_ (heap);
	      return;		/* new item should already be set in heap */
	    }
	  ASSERT (heap->hz < heap->maxsz);
	  heap->loc[HTSIZE + heap->hz] = id;
	  heap->hz++;
	}
#else
      BKHEAP_COLLISION (heap);
	/** add to locator overflow */
      if (heap->hz == heap->maxsz)
	{
	  loc_compact_overfow_ (heap);
	  return;		/* new item should already be set in heap */
	}
      ASSERT (heap->hz < heap->maxsz);
      heap->loc[HTSIZE + heap->hz] = id;
      heap->hz++;
#endif
    }
}

/**
 * Check whether item is set in the locator hash
 */
inline int
loc_isset_ (bkheap * const heap, const idx_t id)
{
  size_t i;
  uint h;

  BKHEAP_LOOKUP (heap);

  h = loc_h1_ (id);
  if (BKHEAP_NKEY == heap->loc[h])
    {
      return 0;
    }
  else if (id == heap->loc[h])
    {
      return 1;
    }
  else
    {
#ifdef HTHASHD
      h = loc_h2_ (id);
      if (BKHEAP_NKEY == heap->loc[h])
	{
	  return 0;
	}
      else if (id == heap->loc[h])
	{
	  return 1;
	}
      else
	{
	  BKHEAP_COLLISION (heap);
	  for (i = 0; i < heap->hz; ++i)
	    {
	      if (id == heap->loc[HTSIZE + i])
		{
		  return 1;
		}
	    }
	  return 0;
	}
#else
      BKHEAP_COLLISION (heap);
      for (i = 0; i < heap->hz; ++i)
	{
	  if (id == heap->loc[HTSIZE + i])
	    {
	      return 1;
	    }
	}
      return 0;
#endif
    }
  return 0;
}


/**
 * Remove an item from the locator hash
 */
inline void
loc_unset_ (bkheap * const heap, const idx_t id)
{
  size_t i;
  uint h;

  BKHEAP_LOOKUP (heap);

  h = loc_h1_ (id);
#ifdef HTHASHD
  if (id == heap->loc[h])
    {
      heap->loc[h] = BKHEAP_SKEY;
    }
  else
    {
      h = loc_h2_ (id);
      if (id == heap->loc[h])
	{
	  heap->loc[h] = BKHEAP_SKEY + (heap->hz == 0);	/* Note SKEY+1 = NKEY */
	}
      else
	{
	  BKHEAP_COLLISION (heap);
	    /** check the locator overflow */
	  for (i = 0; i < heap->hz; ++i)
	    {
	      if (id == heap->loc[HTSIZE + i])
		{
		  if (i == heap->hz - 1)
		    {
		      heap->hz--;
		    }
		  else
		    {
		      heap->loc[HTSIZE + i] = BKHEAP_SKEY;
		    }
		  return;
		}
	    }
	}
    }
#else
  if (id == heap->loc[h])
    {
      heap->loc[h] = BKHEAP_SKEY + (heap->hz == 0);	/* Note SKEY+1 = NKEY */
    }
  else
    {
      BKHEAP_COLLISION (heap);
	/** check the locator overflow */
      for (i = 0; i < heap->hz; ++i)
	{
	  if (id == heap->loc[HTSIZE + i])
	    {
	      if (i == heap->hz - 1)
		{
		  heap->hz--;
		}
	      else
		{
		  heap->loc[HTSIZE + i] = BKHEAP_SKEY;
		}
	      return;
	    }
	}
    }
#endif
}

/**
 * Remove an item from the locator hash, ignoring the locator overflow
 */
inline void
loc_unset_shallow_ (bkheap * const heap, const idx_t id)
{
  size_t i;
  uint h;

  BKHEAP_LOOKUP (heap);

  h = loc_h1_ (id);
#ifdef HTHASHD
  if (id == heap->loc[h])
    {
      heap->loc[h] = BKHEAP_SKEY + (heap->hz == 0);	/* Note SKEY+1 = NKEY */
    }
  else
    {
      BKHEAP_COLLISION (heap);
      h = loc_h2_ (id);
      if (id == heap->loc[h])
	{
	  heap->loc[h] = BKHEAP_SKEY;
	}
    }
#else
  if (id == heap->loc[h])
    {
      heap->loc[h] = BKHEAP_SKEY + (heap->hz == 0);	/* Note SKEY+1 = NKEY */
      return;
    }
  BKHEAP_COLLISION (heap);
#endif
}


/**
 * Reset the locator hash
 */
inline void
loc_reset_ (bkheap * const heap)
{
  size_t i;

  heap->hz = 0;
  for (i = 0; i < HTSIZE; ++i)
    {
      heap->loc[i] = BKHEAP_NKEY;
    }
  for (i = 0; i < heap->sz; ++i)
    {
      loc_set_ (heap, heap->data[i].key);
    }
}




/** PUBLIC FUNCTIONS **/



/**
 * Find the min of values currently in the heap
 */
inline val_t
bkheapMin (const bkheap * const heap)
{
  return heap->sz > 0 ? heap->data[0].val : BKHEAP_NVAL;
};

/**
 * Find the min of values currently in the heap
 */
inline idx_t
bkheapMinKey (const bkheap * const heap)
{
  return heap->sz > 0 ? heap->data[0].key : BKHEAP_NKEY;
};



/**
 * Create a heap structure
 */
bkheap *
bkheapCreate (const size_t maxsz,
	      const size_t maxosz, const char createLocator)
{

  bkheap *h = NULL;

  if (maxsz < 1)
    {
      da_errexit ("bkheapCreate: maxsz must be greater than 0.");
    }
  if (maxosz < 0)
    {
      da_errexit ("bkheapCreate: maxosz must be greater or equal to 0.");
    }

  h = (bkheap *) da_malloc (sizeof (bkheap), "bkheapCreate: h");
  memset (h, 0, sizeof (bkheap));
  h->maxsz = maxsz;
  h->maxosz = maxosz;
  h->duplicates = 1;
  h->data = da_ivkvsmalloc (maxsz + maxosz, (da_ivkv_t)
			    {
			    BKHEAP_NKEY, BKHEAP_NVAL}
			    , "bkheapCreate: h->heap");
  h->loc = NULL;
  if (createLocator)
    {
      h->loc =
	da_ismalloc (HTSIZE + maxsz, BKHEAP_NKEY, "bkheapCreate: h->loc");
    }

  return h;
}


/**
 * Free a given heap
 */
void
bkheapFree (bkheap ** heap)
{
  bkheap *h = *heap;

  if (h->loc)
    {
      da_free ((void **) &h->loc, LTERM);
    }
  da_free ((void **) &h->data, heap, LTERM);
}


/**
 * Empty the data structure
 */
void
bkheapClear (bkheap * const heap)
{
  da_ivkvset (heap->maxsz, (da_ivkv_t)
	      {
	      BKHEAP_NKEY, BKHEAP_NVAL}, heap->data);
  da_iset (HTSIZE, BKHEAP_NKEY, heap->loc);
  heap->hz = 0;
  heap->sz = 0;
  heap->osz = 0;
}

/**
 * Completely empty the data structure + reset counters
 */
void
bkheapReset (bkheap * const heap)
{
  bkheapClear (heap);
  heap->hc = 0;
  heap->hl = 0;
}

/**
 * Copy the contents of one heap into another of same size
 */
void
bkheapCopy (bkheap * const from, bkheap * const to)
{
  ASSERT (from && to);
  ASSERT (from != to);
  if (from->maxsz != to->maxsz || from->maxosz != to->maxosz)
    {
      da_errexit ("bkheapCopy: from and to do not have the same size.");
    }
  memcpy (to->data, from->data,
	  sizeof (da_ivkv_t) * (from->maxsz + from->maxosz));
  memcpy (to->loc, from->loc, sizeof (idx_t) * (HTSIZE + from->maxsz));
  to->sz = from->sz;
  to->osz = from->osz;
  to->hc = from->hc;
  to->hl = from->hl;
  to->hz = from->hz;
  to->duplicates = from->duplicates;
}


/**
 * Insert an item directly into the heap, ignoring the overflow. Item is guaranteed to be new and
 * need not be checked to see if it already exists in the heap.
 */
int
bkheapInsertHeapNew (bkheap * const heap, const idx_t id, const val_t val)
{
  ptr_t i, j, full;
  val_t v;
  da_ivkv_t *data, d;

  if (((full = heap->sz == heap->maxsz) && val <= bkheapMin (heap)))
    {
      return 0;
    }

  data = heap->data;

  if (full)
    {
	/** heap is full; remove the top and heapify */
      loc_unset_ (heap, data[0].key);
      heap->sz--;
      if (heap->sz > 0)
	{
	  d = data[heap->sz];
	  v = d.val;
	  i = 0;
	  while ((j = (i << 1) + 1) < heap->sz)
	    {
	      if (data[j].val < v)
		{
		  if (j + 1 < heap->sz && data[j + 1].val < data[j].val)
		    {
		      j++;
		    }
		  data[i] = data[j];
		  i = j;
		}
	      else if (j + 1 < heap->sz && data[j + 1].val < v)
		{
		  j++;
		  data[i] = data[j];
		  i = j;
		}
	      else
		{
		  break;
		}
	    }
	  data[i] = d;
	}
    }

    /** insert new element */
  i = heap->sz++;
  while (i > 0)
    {
      j = PARENT (i);
      if (val < heap->data[j].val)
	{
	  heap->data[i] = heap->data[j];
	  i = j;
	}
      else
	break;
    }
  ASSERT (i >= 0);
  heap->data[i].val = val;
  heap->data[i].key = id;

  loc_set_ (heap, id);

  return 1;
}


/**
 * Insert an item directly into the heap, ignoring the overflow.
 */
int
bkheapInsertHeap (bkheap * const heap, const idx_t id, const val_t val)
{
  if (heap->duplicates && loc_isset_ (heap, id))
    {
      return 0;
    }
  return bkheapInsertHeapNew (heap, id, val);
}


/**
 * Insert an item into the heap if it can improve heap. If heap if full, insert in overflow. If overflow is full,
 * compact heap first (insert items from overflow into heap).
 */
int
bkheapInsert (bkheap * const heap, const idx_t id, const val_t val)
{
  ptr_t n;
  val_t v;

  if (heap->sz < heap->maxsz || heap->maxosz == 0)
    {
      return bkheapInsertHeap (heap, id, val);
    }

  if (val <= bkheapMin (heap) || (heap->duplicates && loc_isset_ (heap, id)))
    {
      return 0;
    }

  n = 0;
  if (heap->osz == heap->maxosz)
    {
      n += bkheapCompact (heap);
    }

  ASSERT (heap->osz < heap->maxosz);
  heap->data[heap->maxsz + heap->osz] = (da_ivkv_t)
  {
  id, val};
  heap->osz++;

  return n + 1;
}


/**
 *
 * @param cand List of candidates to select top-k from
 * @param k Number of top-k items to select
 * @param lo First index in cand to select from
 * @param len Length of section to select from
 */
void
bkheap_select_ (da_ivkv_t * const cand, const ptr_t k, ptr_t lo, ptr_t len)
{
  ptr_t i, j, mid, hi;
  val_t v;
  da_ivkv_t d;

  hi = lo + len - 1;

  for (; lo < hi;)
    {
      mid = lo + ((hi - lo) >> 1);

      /* select the median */
      if (cand[lo].val < cand[mid].val)
	mid = lo;
      if (cand[hi].val > cand[mid].val)
	mid = hi;
      else
	goto jump_over;
      if (cand[lo].val < cand[mid].val)
	mid = lo;

    jump_over:
      BKHEAP_SWAP (cand[mid], cand[hi], d);
      v = cand[hi].val;

      /* the partitioning algorithm */
      for (i = lo - 1, j = lo; j < hi; j++)
	{
	  if (cand[j].val >= v)
	    {
	      i++;
	      BKHEAP_SWAP (cand[i], cand[j], d);
	    }
	}
      i++;
      BKHEAP_SWAP (cand[i], cand[hi], d);

      if (i > k)
	{
	  hi = i - 1;
	}
      else if (i < k)
	{
	  lo = i + 1;
	}
      else
	{
	  break;
	}
    }
}


void
heapify_ (bkheap * const heap, ptr_t i)
{
  ptr_t j;
  val_t v;
  da_ivkv_t *data, d;

  if (i >= heap->sz || (ssize_t) i < 0)
    {
      return;
    }

  data = heap->data;
  d = data[i];
  v = d.val;
  while ((j = (i << 1) + 1) < heap->sz)
    {
      if (data[j].val < v)
	{
	  if (j + 1 < heap->sz && data[j + 1].val < data[j].val)
	    {
	      j++;
	    }
	  data[i] = data[j];
	  i = j;
	}
      else if (j + 1 < heap->sz && data[j + 1].val < v)
	{
	  j++;
	  data[i] = data[j];
	  i = j;
	}
      else
	{
	  break;
	}
    }
  data[i] = d;
}

/**
 * Enter the overflow items into the heap. Overflow items are all assumed greater or equal than heap min.
 */
int
bkheapCompact (bkheap * const heap)
{

  ptr_t i, j, n, lo, mid, hi;
  da_ivkv_t *cand, *data, d;

  if (heap->osz == 0 || heap->maxosz == 0)
    {				/* no overflow */
      return 0;
    }

  data = heap->data;
  cand = heap->data + heap->maxsz;
  n = 0;

  while (heap->sz < heap->maxsz && heap->osz > 0)
    {				/* fill out the heap proper first */
      heap->osz--;
      n += bkheapInsertHeap (heap, cand[heap->osz].key, cand[heap->osz].val);
    }

  if (heap->osz < heap->maxsz)
    {				/* fewer than heap-size items to insert */
      for (i = 0; i < heap->osz; ++i)
	{
	  n += bkheapInsertHeap (heap, cand[i].key, cand[i].val);
	}
      heap->osz = 0;
      return n;
    }

  if (heap->duplicates)
    {
      /* check for and remove duplicates */
      for (i = 0; i < heap->osz; ++i)
	{
	  while (loc_isset_ (heap, cand[i].key) && heap->osz > i)
	    {			/* duplicate item */
	      heap->osz--;
	      BKHEAP_SWAP (cand[i], cand[heap->osz], d);
	    }
	  loc_set_ (heap, cand[i].key);
	}

      if (heap->osz < heap->maxsz)
	{
	  loc_reset_ (heap);
	  return n + bkheapCompact (heap);
	}
    }

  /* select top-k of the overflow items, where k = heap-size */
  bkheap_select_ (data, heap->maxsz, 0, heap->maxsz + heap->osz);
  heap->osz = 0;
  bkheapBuildHeap (heap);

  return n + heap->maxsz;

}


/**
 * Given the set of items in [0, heap->sz), make them into a heap
 */
void
bkheapBuildHeap (bkheap * heap)
{
  size_t i;

  for (i = PARENT (heap->sz); i-- > 0 && i < heap->sz;)
    {
      heapify_ (heap, i);
    }

  loc_reset_ (heap);
}


/**
 * Given the set of items in [0, heap->sz), make them into a heap.
 * Do not reset the locator array
 */
void
bkheapBuildHeapNoReset (bkheap * heap)
{
  size_t i;

  for (i = PARENT (heap->sz); i-- > 0 && i < heap->sz;)
    {
      heapify_ (heap, i);
    }
}

/**
 * Print the heap structure
 */
void
bkheapPrint (const bkheap * const heap)
{
  ptr_t i;
  printf
    ("BKHEAP: maxsz %zu, maxosz %zu, sz %zu, osz %zu, hz %zu, hc %zu, hl %zu\n",
     heap->maxsz, heap->maxosz, heap->sz, heap->osz, heap->hz, heap->hc,
     heap->hl);
  printf ("heap: ");
  for (i = 0; i < heap->sz; ++i)
    {
      printf (PRNT_IDXTYPE " %g, ", heap->data[i].key, heap->data[i].val);
    }
  printf ("\n");
  fflush (stdout);
  printf ("overflow: ");
  for (i = heap->maxsz; i < heap->maxsz + heap->osz; ++i)
    {
      printf (PRNT_IDXTYPE " %g, ", heap->data[i].key, heap->data[i].val);
    }
  printf ("\n");
  fflush (stdout);
  printf ("locator: ");
  for (i = 0; i < HTSIZE; ++i)
    {
      if (heap->loc[i] == BKHEAP_NKEY)
	{
	  continue;
	}
      if (heap->loc[i] == BKHEAP_SKEY)
	{
	  printf ("_" PRNT_PTRTYPE "_ ", i);
	}
      else
	{
	  printf (PRNT_PTRTYPE " " PRNT_IDXTYPE ", ", i, heap->loc[i]);
	}
    }
  printf ("\n");
  fflush (stdout);
  printf ("locator overflow: ");
  for (; i < HTSIZE + heap->maxsz; ++i)
    {
      if (heap->loc[i] == BKHEAP_NKEY)
	{
	  continue;
	}
      if (heap->loc[i] == BKHEAP_SKEY)
	{
	  printf ("_" PRNT_PTRTYPE "_ ", i);
	}
      else
	{
	  printf (PRNT_PTRTYPE " " PRNT_IDXTYPE ", ", i, heap->loc[i]);
	}
    }
  printf ("\n\n");
  fflush (stdout);
}


/**
 * Associate a locator with the heap. The locator will be reset.
 */
void
bkheapSetLocator (bkheap * const heap, idx_t * loc)
{
  heap->loc = loc;
  loc_reset_ (heap);
}


/**
 * Reset the locator associated with the heap.
 */
void
bkheapResetLocator (bkheap * const heap)
{
  loc_reset_ (heap);
}

/**
 * Check whether an item exists in the heap proper
 */
int
bkheapExists (bkheap * const heap, const idx_t id)
{
  return loc_isset_ (heap, id);
}


/**
 * Check the consistency of the heap and uniqueness of heap keys
 */
int
bkheapCheck (bkheap * const heap)
{
  ptr_t i, j;
  idx_t *keys;

  if (!heap)
    {
      da_errexit ("bkheapCheck: Heap not initialized!");
    }

  if (heap->sz == 0)
    {
      return 1;
    }

  ASSERT (loc_isset_ (heap, heap->data[0].key));
  keys = da_imalloc (heap->sz, "bkheapCheck: keys");
  keys[0] = heap->data[0].key;
  for (i = 1; i < heap->sz; ++i)
    {
      if (heap->data[i].key < BKHEAP_SKEY)
	{
	  ASSERT (loc_isset_ (heap, heap->data[i].key));
	}
      ASSERT (heap->data[PARENT (i)].val < heap->data[i].val);
      ASSERT (heap->data[0].val < heap->data[i].val);
      keys[i] = heap->data[i].key;
    }
  /* check keys for uniqueness */
  da_isorti (heap->sz, keys);
  for (i = 1; i < heap->sz; ++i)
    {
      ASSERT (keys[i] != keys[i - 1]);
    }
  da_free ((void **) &keys, LTERM);
  /* check that overflow values are greater or equal than heap min */
  if (heap->maxsz > heap->sz)
    {
      ASSERT (heap->osz == 0);
    }

  return 1;
}
