/*
 * bkheap.h
 *
 *  Min-bounded k-select heap data structure
 *
 *  The data structure is a combination of a binary heap and some overflow space which is used to minimize
 *  random accesses in the heap data structure. The structure is optimized for top-k identification via
 *  multiple inserts, does not support removal or update of elements, and works as follows:
 *  - When less than k elements exist, min=MIN and elements are simply added in the order they arrive
 *  - When the k-th element is added, the first k elements are heapified
 *  - Additional elements are only added to the overflow if they are >= min. When the overflow is full,
 *    k-select is called on the overflow (if overflow size > k) and the k elements are inserted into the heap,
 *    each replacing current min if greater than the current min.
 *
 *  Created on: Mar 22, 2016
 *      Author: David C. Anastasiu
 */

#ifndef BKHEAP_H_
#define BKHEAP_H_

#include "includes.h"

#define BKHEAP_NVAL VAL_MIN	/* null value */
#define BKHEAP_NKEY IDX_MAX	/* null key */
#define BKHEAP_SKEY IDX_MAX-1	/* skip key; key was deleted */
#define LCHILD(x)   2*x + 1
#define RCHILD(x)   2*x + 2
#define PARENT(x)   (x-1)>>1
#ifndef HTLEN
#define HTLEN 13
#endif
#define HTSIZE      (1<<HTLEN)	/* Length of hash table array */
#define HTMASK      (HTSIZE-1)	/* Mask for hash table array index */
//#define HTHASHD  /* use double-hashing */
/* Byte-wise swap two items. */
#define BKHEAP_SWAP(a, b, tmp) { tmp = (a); (a) = (b); (b) = tmp; }
#ifdef EXTRACOUNTS
#define BKHEAP_COLLISION(heap) heap->hc++;
#define BKHEAP_LOOKUP(heap) heap->hl++;
#else
#define BKHEAP_COLLISION(heap)
#define BKHEAP_LOOKUP(heap)
#endif


/**
 * Create a heap structure
 */
bkheap *bkheapCreate (const size_t maxsz,
		      const size_t maxosz, const char createLocator);

/**
 * Free a given heap
 */
void bkheapFree (bkheap ** heap);

/**
 * Empty the data structure without resetting overflow data, just its count
 */
void bkheapClear (bkheap * const heap);

/**
 * Completely empty the data structure + reset counters
 */
void bkheapReset (bkheap * const heap);

/**
 * Copy the contents of one heap into another of same size
 */
void bkheapCopy (bkheap * const from, bkheap * const to);

/**
 * Find the min of values currently in the heap
 */
val_t bkheapMin (const bkheap * const heap);

/**
 * Find the min of values currently in the heap
 */
idx_t bkheapMinKey (const bkheap * const heap);

/**
 * Insert an item into the heap. If heap if full, insert in overflow. If overflow is full,
 * compact heap (insert items from overflow into heap).
 * Note: locator must be initialized in case overflow becomes full and items must be inserted into heap
 */
int bkheapInsert (bkheap * const heap, const idx_t id, const val_t val);

/**
 * Insert an item directly into the heap, ignoring the overflow.
 * Note: locator must be initialized
 */
int bkheapInsertHeap (bkheap * const heap, const idx_t id, const val_t val);

/**
 * Insert an item directly into the heap, ignoring the overflow. Item is guaranteed to be new and
 * must not be checked to see if already in the heap.
 */
int bkheapInsertHeapNew (bkheap * const heap,
			 const idx_t id, const val_t val);

/**
 * Given the set of items in [0, heap->sz), make them into a heap
 */
void bkheapBuildHeap (bkheap * heap);

/**
 * Given the set of items in [0, heap->sz), make them into a heap.
 * Do not reset the locator array
 */
void bkheapBuildHeapNoReset (bkheap * heap);

/**
 * Associate a locator with the heap. The locator will be reset.
 */
void bkheapSetLocator (bkheap * const heap, idx_t * loc);

/**
 * Reset the locator associated with the heap.
 */
void bkheapResetLocator (bkheap * const heap);

/**
 * Enter the overflow items into the heap
 */
int bkheapCompact (bkheap * const heap);

/**
 * Check whether an item exists in the heap proper
 */
int bkheapExists (bkheap * const heap, const idx_t id);

/**
 * Print the heap structure
 */
void bkheapPrint (const bkheap * const heap);

/**
 * Check the consistency of the heap
 */
int bkheapCheck (bkheap * const heap);



#endif /* BKHEAP_H_ */
