/*!
 \file  pl2knn.c
 \brief This file contains pL2Knn related functions

 \author David C. Anastasiu
 */

#include "includes.h"
#include "sdhash.h"
#include "pl2knn.h"




// inline functions

/* dot product between sparse and dense vector */
inline accum_t
psddotp (size_t const sz,
	 idx_t const *restrict const ind,
	 val_t const *restrict const val, val_t const *restrict const qrow)
{
  ptr_t register i;
  accum_t register sim;

  for (sim = 0.0, i = 0; i < sz; ++i)
    {
      sim += val[i] * qrow[ind[i]];
    }

  return sim;
}

/**
 * Dot product between sparse and dense vector with pruning,
 * assuming vectors are l2-normalized non-negative vectors.
 * We limit the testing to the top values, since val is sorted
 * in decreasing value and the tail will produce little benefit.
 **/
inline accum_t
psdpdotp (val_t const msim,
	  size_t const sz,
	  idx_t const *restrict const ind,
	  val_t const *restrict const val, val_t const *restrict const qrow)
{
  ptr_t register i;
  accum_t sim = 0.0;
  accum_t csn = 1.0;
  accum_t qsn = 1.0;

  if (sz < PL2KNNA_PRUNE_SZ)
    {

      for (i = 0; i < sz; ++i)
	{
	  csn -= val[i] * val[i];
	  if (qrow[ind[i]] > 0)
	    {
	      qsn -= qrow[ind[i]] * qrow[ind[i]];
	      sim += val[i] * qrow[ind[i]];
	      if (sim + sqrt (csn) * sqrt (qsn) < msim)
		{
		  return -1.0;	/* pruned */
		}
	    }
	}

    }
  else
    {

      for (i = 0; i < PL2KNNA_PRUNE_SZ; ++i)
	{
	  csn -= val[i] * val[i];
	  if (qrow[ind[i]] > 0)
	    {
	      qsn -= qrow[ind[i]] * qrow[ind[i]];
	      sim += val[i] * qrow[ind[i]];
	      if (sim + sqrt (csn) * sqrt (qsn) < msim)
		{
		  return -1.0;	/* pruned */
		}
	    }
	}


      if (sim + sqrt (csn) < msim)
	{
	  return -1.0;		/* pruned */
	}

	/** compute the remainder of the dot-product */
      for (; i < sz; ++i)
	{
	  sim += val[i] * qrow[ind[i]];
	}

    }

  return sim;
}




/**
 * Dot product between sparse and dense vector with pruning,
 * assuming vectors are l2-normalized non-negative vectors.
 * This version of the psdpdotp function uses cached query lengths.
 * Note that the feature processing of the two vectors has to be in
 * the same order, so the original matrix (or a permutation of it)
 * must be used to compute the dot-product. In the psdpdotp function,
 * the feature processing order is irrelevant.
 **/
inline accum_t
psdpdotp2 (val_t const msim,
	   size_t const sz,
	   idx_t const *restrict const ind,
	   val_t const *restrict const val,
	   val_t const *restrict const len,
	   val_t const *restrict const qrow, val_t const *restrict const qrln)
{
  ptr_t register i;
  accum_t sim = 0.0;
  idx_t j;
  val_t v;

  if (sz < PL2KNNA_PRUNE_SZ)
    {

      for (i = 0; i < sz; ++i)
	{
	  j = ind[i];
	  if (qrow[j] > 0)
	    {
	      sim += val[i] * qrow[j];
	      if (sim + len[i] * qrln[j] < msim)
		{
		  return -1.0;	/* pruned */
		}
	    }
	}

    }
  else
    {

      for (i = 0; i < PL2KNNA_PRUNE_SZ; ++i)
	{
	  j = ind[i];
	  if (qrow[j] > 0)
	    {
	      sim += val[i] * qrow[j];
	      if (sim + len[i] * qrln[j] < msim)
		{
		  return -1.0;	/* pruned */
		}
	    }
	}

      if (sim + len[i] < msim)
	{
	  return -1.0;		/* pruned */
	}

	/** compute the remainder of the dot-product */
      for (; i < sz; ++i)
	{
	  sim += val[i] * qrow[ind[i]];
	}

    }

  return sim;
}


/**
 * Version of psdpdotp that only computes l2-norm of the candidate
 */
inline accum_t
psdpdotp3 (val_t const msim,
	   size_t const sz,
	   idx_t const *restrict const ind,
	   val_t const *restrict const val, val_t const *restrict const qrow)
{
  ptr_t register i;
  accum_t sim = 0.0;
  accum_t csn = 1.0;

  if (sz < PL2KNNA_PRUNE_SZ)
    {

      for (i = 0; i < sz; ++i)
	{
	  sim += val[i] * qrow[ind[i]];
	  csn -= val[i] * val[i];
	  if (sim + sqrt (csn) < msim)
	    {
	      return -1.0;	/* pruned */
	    }
	}

    }
  else
    {

      for (i = 0; i < PL2KNNA_PRUNE_SZ; ++i)
	{
	  sim += val[i] * qrow[ind[i]];
	  csn -= val[i] * val[i];
	  if (sim + sqrt (csn) < msim)
	    {
	      return -1.0;	/* pruned */
	    }
	}

	/** compute the remainder of the dot-product */
      for (; i < sz; ++i)
	{
	  sim += val[i] * qrow[ind[i]];
	}

    }

  return sim;
}

/* dot product between sparse and sparse vector with packing */
inline accum_t
pssdotp (size_t const csz,
	 idx_t const *restrict const cind,
	 val_t const *restrict const cval,
	 size_t const qsz,
	 idx_t const *restrict const qind,
	 val_t const *restrict const qval, val_t * restrict const tmp)
{

  ptr_t register i;
  ptr_t register iq;
  ptr_t register ic;
  accum_t register sim;

  iq = ic = 0;
  if (qsz > csz)
    {				/* packing query down to the size of the candidate */
      while (iq < qsz && ic < csz)
	{
	  if (qind[iq] < cind[ic])
	    {
	      iq++;
	    }
	  else if (qind[iq] > cind[ic])
	    {
	      tmp[ic] = 0.0;
	      ic++;
	    }
	  else
	    {
	      tmp[ic] = qval[iq];
	      iq++;
	      ic++;
	    }
	}
      while (ic < csz)
	{
	  tmp[ic++] = 0.0;
	}

      for (sim = 0.0, i = 0; i < csz; ++i)
	{
	  sim += tmp[i] * cval[i];
	}
    }
  else
    {				/* packing candidate down to the size of the query */
      while (iq < qsz && ic < csz)
	{
	  if (qind[iq] < cind[ic])
	    {
	      tmp[iq] = 0.0;
	      iq++;
	    }
	  else if (qind[iq] > cind[ic])
	    {
	      ic++;
	    }
	  else
	    {
	      tmp[iq] = cval[ic];
	      iq++;
	      ic++;
	    }
	}
      while (iq < qsz)
	{
	  tmp[iq++] = 0.0;
	}

      for (sim = 0.0, i = 0; i < qsz; ++i)
	{
	  sim += tmp[i] * qval[i];
	}
    }

  return sim;
}


/* sparse sparse dot-product */
inline accum_t
ssdotp (size_t const csz,
	idx_t const *restrict const cind,
	val_t const *restrict const cval,
	size_t const qsz,
	idx_t const *restrict const qind, val_t const *restrict const qval)
{

  ptr_t register iq;
  ptr_t register ic;
  accum_t register sim;

  iq = ic = 0;
  sim = 0.0;
  while (iq < qsz && ic < csz)
    {
      if (qind[iq] < cind[ic])
	{
	  iq++;
	}
      else if (qind[iq] > cind[ic])
	{
	  ic++;
	}
      else
	{
	  sim += qval[iq] * cval[ic];
	  iq++;
	  ic++;
	}
    }

  return sim;
}


inline void
pl2knnTransferCounts (params_t * const params,
		      da_tsearch_t * const *const search)
{
  size_t i;
  for (i = 0; i < params->nthreads; ++i)
    {
      params->nCandidates += search[i]->nCandidates;
      params->nDotProducts += search[i]->nDotProducts;
      params->nDotProducts1 += search[i]->nDotProducts1;
      params->nDotProducts2 += search[i]->nDotProducts2;
      params->nPruneDotP += search[i]->nPruneDotP;
      params->nPruneDotP2 += search[i]->nPruneDotP2;
      params->nPruneLength += search[i]->nPruneLength;
      params->nPruneLength2 += search[i]->nPruneLength2;
      params->nPruneMinsize += search[i]->nPruneMinsize;
      params->nPrunePscore += search[i]->nPrunePscore;
      params->nHeapInserts += search[i]->nHeapInserts;
    }
}



void
pl2knnPrintCompileOptions ()
{

  printf ("pL2knn-a compile options:\n");

#if defined(PL2KNNA_PRESELECT) && PL2KNNA_PRESELECT
  printf ("\tpreselect:              true\n");
#else
  printf ("\tpreselect:              false\n");
#endif

#if defined(PL2KNNA_PRUNE) && PL2KNNA_PRUNE
  printf ("\tl2-norm prune:          true\n");
  printf ("\tprune min:              %g\n", PL2KNNA_PRUNE_MIN);
  printf ("\tprune size:             %d\n", PL2KNNA_PRUNE_SZ);
#if defined(Pl2KNNA_HINSERT) && Pl2KNNA_HINSERT
  printf ("\tdirect heap insert:     true\n");
#else
  printf ("\tdirect heap insert:     false\n");
#endif
#if PL2KNNA_PRUNE_NCACHE
  printf ("\tcache pruning l2-norms: true\n");
#else
  printf ("\tcache pruning l2-norms: false\n");
#endif
#else
  printf ("\tl2-norm prune:          false\n");
#endif

#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  printf ("\tqhash:                  dense\n");
#else
  printf ("\tqhash:                  hash-table\n");

#if PL2KNN_QHASH_HTSIZE > 0
  printf ("\thtsize choice:          present\n");
#else
  printf ("\thtsize choice:          auto\n");
#endif
#endif

  fflush (stdout);
}




da_csr_t *
pl2knnCreateEnhanceIndex (idx_t const nn,
			  idx_t const nrows,
			  da_iapq_t * const *const knng,
			  da_tsearch_t * const *const search)
{
  ssize_t ii;

  const int nt = omp_get_max_threads ();
  da_csr_t *const mat = da_csr_Create ();
  mat->nrows = mat->ncols = nrows;
  mat->rowptr = da_pnmalloc_a (nrows + 1, "pl2knnCreateEnhanceIndex: fptr");

#pragma omp parallel default(none)
  {
    idx_t register i;
    ptr_t register j;
    ptr_t register k;
    ptr_t s, e, cs, nnz;

    const int tid = omp_get_thread_num ();
    const idx_t nr = mat->nrows;
    ptr_t *const fptr = mat->rowptr;
    const idx_t bsz = ceil (nr / (double) nt);
    da_ivkv_t *restrict const cand = search[tid]->scands;

    /* find nnz and populate forward pointer */
#pragma omp for schedule(static)
    for (i = 0; i < nrows; ++i)
      {
	fptr[i] = knng[i]->nnodes;
      }

#pragma omp barrier
#pragma omp single
    {
      CSRMAKE (i, nr, fptr);
      nnz = fptr[nr];
      mat->rowind = da_imalloc_a (nnz, "pl2knnCreateEnhanceIndex: find");
      mat->rowval = da_vmalloc_a (nnz, "pl2knnCreateEnhanceIndex: fval");
    }

    idx_t *const find = mat->rowind;
    val_t *const fval = mat->rowval;

    /* sort and transfer assigned section of knng values into row structure */
#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC_FAST)
    for (i = 0; i < nrows; ++i)
      {
	da_iapq_t const *restrict const kg = knng[i];
	da_iakv_t const *restrict const hp = knng[i]->heap;
	for (j = 0; j < kg->nnodes; ++j)
	  {
	    cand[j].key = hp[j].key;
	    cand[j].val = hp[j].val;
	  }
	da_ivkvsortd (kg->nnodes, cand);
	for (j = 0; j < kg->nnodes; ++j)
	  {
	    find[fptr[i]] = cand[j].key;
	    fval[fptr[i]++] = cand[j].val;
	  }
      }

  }

  CSRSHIFT (ii, nrows, mat->rowptr);

  return mat;

}

/**
 * Version of pl2knnCreateEnhanceIndex that also creates inverted neighborhood index
 */
da_csr_t *
pl2knnCreateEnhanceIndex2 (idx_t const nn,
			   idx_t const nrows,
			   da_iapq_t * const *const knng,
			   da_tsearch_t * const *const search)
{
  ssize_t ii;

  const int nt = omp_get_max_threads ();
  da_csr_t **const tmps =
    (da_csr_t **) da_malloc (sizeof (da_csr_t *) * nt,
			     "pl2knnCreateEnhanceIndex2: tmps");
  for (ii = 0; ii < nt; ++ii)
    {
      tmps[ii] = NULL;
    }
  da_csr_t *const mat = da_csr_Create ();
  mat->nrows = mat->ncols = nrows;
  mat->rowptr = da_pnmalloc_a (nrows + 1, "pl2knnCreateEnhanceIndex2: fptr");
  mat->colptr = da_pnmalloc_a (nrows + 1, "pl2knnCreateEnhanceIndex2: rptr");

#pragma omp parallel default(none)
  {
    idx_t register i;
    ptr_t register j;
    ptr_t register k;
    ptr_t s, e, cs, nnz;

    const int tid = omp_get_thread_num ();
    const idx_t nr = mat->nrows;
    ptr_t *const fptr = mat->rowptr;
    ptr_t *const rptr = mat->colptr;
    const idx_t bsz = ceil (nr / (double) nt);
    da_ivkv_t *restrict const cand = search[tid]->scands;

    /* find nnz and populate forward pointer */
#pragma omp for schedule(static)
    for (i = 0; i < nrows; ++i)
      {
	fptr[i] = knng[i]->nnodes;
      }

#pragma omp barrier
#pragma omp single
    {
      CSRMAKE (i, nr, fptr);
      nnz = fptr[nr];
      mat->rowind = da_imalloc_a (nnz, "pl2knnCreateEnhanceIndex2: find");
      mat->rowval = da_vmalloc_a (nnz, "pl2knnCreateEnhanceIndex2: fval");
      mat->colind = da_imalloc_a (nnz, "pl2knnCreateEnhanceIndex2: rind");
      mat->colval = da_vmalloc_a (nnz, "pl2knnCreateEnhanceIndex2: rval");
    }

    idx_t *const find = mat->rowind;
    val_t *const fval = mat->rowval;
    idx_t *const rind = mat->colind;
    val_t *const rval = mat->colval;

    /* create the temporary assigned tile of the input matrix */
    da_csr_t *const tmp = tmps[tid] = da_csr_Create ();
    s = tid * bsz;
    e = da_min ((tid + 1) * bsz, nr);
    tmp->nrows = e - s;
    tmp->ncols = nr;
    tmp->rowind = find;
    tmp->rowval = fval;
    tmp->rowptr = fptr + s;

    /* sort and transfer assigned section of knng values into row structure */
#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC_FAST)
    for (i = 0; i < nrows; ++i)
      {
	da_iapq_t const *restrict const kg = knng[i];
	da_iakv_t const *restrict const hp = knng[i]->heap;
	for (j = 0; j < kg->nnodes; ++j)
	  {
	    cand[j].key = hp[j].key;
	    cand[j].val = hp[j].val;
	  }
	da_ivkvsortd (kg->nnodes, cand);
	for (j = 0; j < kg->nnodes; ++j)
	  {
	    find[fptr[i]] = cand[j].key;
	    fval[fptr[i]++] = cand[j].val;
	  }
      }

#pragma omp barrier
#pragma omp single
    {
      CSRSHIFT (i, nr, fptr);
    }

    /* transpose the matrix with an offset */
    da_csr_CreateOffsetIndex (tmp, DA_COL, s);

#pragma omp barrier

    /* create reverse pointer array */
    for (j = (tid * bsz); j < da_min ((tid + 1) * bsz, nr); ++j)
      {
	for (k = 0, i = 0; i < nt; ++i)
	  {
	    k += tmps[i]->colptr[j + 1] - tmps[i]->colptr[j];
	  }
	rptr[j] = k;
      }

#pragma omp barrier
#pragma omp single
    {
      CSRMAKE (i, nr, rptr);
    }

    /* copy thread indexes into the master inverted index */
    for (j = (tid * bsz); j < da_min ((tid + 1) * bsz, nr); ++j)
      {
	for (i = 0; i < nt; ++i)
	  {
	    cs = tmps[i]->colptr[j];
	    k = tmps[i]->colptr[j + 1] - cs;
	    da_icopy (k, tmps[i]->colind + cs, rind + rptr[j]);
	    da_vcopy (k, tmps[i]->colval + cs, rval + rptr[j]);
	    rptr[j] += k;
	  }
      }

#pragma omp barrier

    /* free tmp memory */
    tmp->rowptr = NULL;
    tmp->rowind = NULL;
    tmp->rowval = NULL;
    tmps[tid] = NULL;
    da_csr_Free ((da_csr_t **) & tmp);
  }

  CSRSHIFT (ii, nrows, mat->colptr);

  da_free ((void **) &tmps, LTERM);

  /* sort the topk values of each inv. list */
  da_pl2knnSortTopKValues (mat, DA_COL, nn, 0, search);


  return mat;

}



/*************************************************************************/
/*! Sorts the topK values (and associated indices) in increasing or decreasing order
    \param mat the matrix itself,
    \param what is either DA_ROW or DA_COL indicating which set of
           values/indices to sort.
    \param topk the number of top values to sort in each row or column
    \param mrcl maximum row/column length (must be expressly provided unless cands has been pre-allocated)
    \param offset ignore the beginning offset items in each row
    \param i_cands optional reusable data structure for thread candidate lists. If passed in, passing
            function responsible for freeing contents
*/
/**************************************************************************/
void
da_pl2knnSortTopKValues (da_csr_t * const mat,
			 char const what,
			 idx_t const topk,
			 idx_t const offset,
			 da_tsearch_t * const *const search)
{
  idx_t n;
  ptr_t *optr;
  idx_t *oind;
  val_t *oval;

  switch (what)
    {
    case DA_ROW:
      if (!mat->rowptr)
	da_errexit
	  ("da_pl2knnSortTopKValues: Row-based view of the matrix does not exists.\n");

      n = mat->nrows;
      optr = mat->rowptr;
      oind = mat->rowind;
      oval = mat->rowval;
      break;

    case DA_COL:
      if (!mat->colptr)
	da_errexit
	  ("da_pl2knnSortTopKValues: Column-based view of the matrix does not exists.\n");

      n = mat->ncols;
      optr = mat->colptr;
      oind = mat->colind;
      oval = mat->colval;
      break;

    default:
      da_errexit ("da_pl2knnSortTopKValues: Invalid index type of %d.\n",
		  what);
      return;
    }

  if (!oval)
    da_errexit
      ("da_pl2knnSortTopKValues: values not present in %s structure of the matrix.",
       what == DA_ROW ? "row" : "col ");

#pragma omp parallel default(none) shared(optr, oind, oval, n)
  {

    idx_t register i;
    ptr_t register j;
    ptr_t register s;
    ptr_t register e;
    ptr_t register l;
    char k;
    ptr_t nk;

    ptr_t *const ptr = optr;
    idx_t *const ind = oind;
    val_t *const val = oval;

    int const tid = omp_get_thread_num ();
    da_ivkv_t *const cand = search[tid]->scands;

#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC_FAST)
    for (i = 0; i < n; ++i)
      {
	s = ptr[i] + offset;
	e = ptr[i + 1];
	l = e - s;
	if (l < 2)
	  {
	    continue;
	  }
	else if (l <= topk)
	  {
	    cand[0].key = ind[s];
	    for (k = 0, j = 1; j < l; ++j)
	      {
		if (val[s + j] > val[s + j - 1])
		  {
		    k = 1;	/* an inversion */
		    break;
		  }
		cand[j].key = ind[s + j];
	      }
	    if (k)
	      {
		for (; j < l; ++j)
		  {
		    cand[j].key = ind[s + j];
		  }
		for (j = 0; j < l; ++j)
		  {
		    cand[j].val = val[s + j];
		  }
		da_ivkvsortd (e - s, cand);
		for (j = 0; j < l; ++j)
		  {
		    ind[s + j] = cand[j].key;
		  }
		for (j = 0; j < l; ++j)
		  {
		    val[s + j] = cand[j].val;
		  }
	      }
	  }
	else
	  {
	    cand[0].key = ind[s];
	    for (k = 0, j = 1; j < l; ++j)
	      {
		if (val[s + j] > val[s + j - 1])
		  {
		    k = 1;	/* an inversion */
		    break;
		  }
		cand[j].key = ind[s + j];
	      }
	    if (k)
	      {
		for (; j < l; ++j)
		  {
		    cand[j].key = ind[s + j];
		  }
		for (j = 0; j < l; ++j)
		  {
		    cand[j].val = val[s + j];
		  }
		nk = da_ivkvkselectd (l, topk, cand);
		da_ivkvsortd (nk, cand);
		for (j = 0; j < l; ++j)
		  {
		    ind[s + j] = cand[j].key;
		  }
		for (j = 0; j < l; ++j)
		  {
		    val[s + j] = cand[j].val;
		  }
	      }
	  }
      }
  }

}




inline size_t
pl2knnCheckCandidates (idx_t const qid,
		       idx_t const start,
		       bool const checkfull,
		       ptr_t const *restrict const ptr,
		       idx_t const *restrict const ind,
		       val_t const *restrict const val,
		       da_iapq_t * restrict const lnn,
		       da_tsearch_t * restrict const ts,
		       size_t const ncands,
		       da_iakv_t * restrict const anbrs,
		       idx_t * const *restrict const cnts)
{

  ptr_t register j;
  ptr_t register n;
  idx_t register k;
  idx_t register cid;
  double register minsim;
  ptr_t nadd;


  idx_t const kk = ts->k;
  idx_t const bsz = ts->bsz;
  idx_t const nthr = ts->nthreads;
  bit *restrict const cbit = ts->cbit;
  da_iakv_t *restrict const qcands = ts->qcands;
  idx_t *restrict const cnt = ts->cnt;
  ptr_t const qsz = ptr[qid + 1] - ptr[qid];
  idx_t const *const qind = ind + ptr[qid];
  val_t const *const qval = val + ptr[qid];
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  val_t *restrict const qrow = ts->qrow;
#else
  sdhash *restrict const qhash = ts->qhash;
#endif
  lnn->locator = ts->loc;
  nadd = 0;


  etimer_start (ts->timer_5);	/* query hash time */
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  for (j = 0; j < qsz; ++j)
    {				/* cache query row */
      qrow[qind[j]] = qval[j];
    }
#else
  sdhashSet (qhash, qind, qval, qsz);
#endif
  etimer_stop (ts->timer_5);	/* query hash time */


  etimer_start (ts->timer_6);	/* dot-product time */
  /* compute dot-products */
  for (n = 0, j = 0; j < ncands; ++j)
    {
      cid = qcands[j].key;
      if (ba_value (cbit, cid))
	{
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
	  qcands[j].val =
	    psddotp (ptr[cid + 1] - ptr[cid], ind + ptr[cid], val + ptr[cid],
		     qrow);
#else
	  qcands[j].val =
	    sdhashDotp (qhash, ind + ptr[cid], val + ptr[cid],
			ptr[cid + 1] - ptr[cid]);
#endif
	  n++;
	}
    }
  ts->nDotProducts += n;
  etimer_stop (ts->timer_6);	/* dot-product time */


  etimer_start (ts->timer_7);	/* local update time */
  da_iapqInitLocator (lnn);
  n = ncands;
#if defined(PL2KNNA_PRESELECT) && PL2KNNA_PRESELECT
  if (ncands > kk)
    {
      n = da_iakvkselectd (ncands, kk, qcands);
    }
#endif
  j = 0;
  if (checkfull)
    {
      while (j < n && lnn->nnodes < kk)
	{
	  if (qcands[j].val > 0)
	    {
	      da_iapqInsert (lnn, qcands[j].key, qcands[j].val);
	      nadd++;
	    }
	  j++;
	}
    }
  minsim = da_iapqSeeTopVal (lnn);
  for (; j < n; ++j)
    {
      cid = qcands[j].key;
      if (qcands[j].val > minsim)
	{
	  da_iapqInsert (lnn, cid, qcands[j].val);
	  minsim = da_iapqSeeTopVal (lnn);
	  nadd++;
	}
    }
  da_iapqResetLocator (lnn);
  etimer_stop (ts->timer_7);	/* local update time */


  etimer_start (ts->timer_5);	/* query hash time */
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  for (j = 0; j < qsz; ++j)
    {
      qrow[qind[j]] = 0.0;	/* reset query row */
    }
#else
  sdhashUnset (qhash, qind, qsz);
#endif
  etimer_stop (ts->timer_5);	/* query hash time */


  /* bucket neighbors for each thread */
  etimer_start (ts->timer_8);	/* shuffle time */
  CSRMAKE (j, nthr, cnt);
  memcpy (cnts[qid - start], cnt, sizeof (idx_t) * (nthr + 1));
  for (j = 0; j < ncands; ++j)
    {
      k = (qcands[j].key / bsz);
      anbrs[cnt[k]++] = qcands[j];
    }
  etimer_stop (ts->timer_8);	/* shuffle time */

  return nadd;

}



/**
 * This version of the pl2knnCheckCandidate uses L2-norm based pruning to forego some of the dot-product computations
 * and stop others early
 */
inline size_t
pl2knnCheckCandidatesPrune (idx_t const qid,
			    idx_t const start,
			    bool const checkfull,
			    ptr_t const *restrict const ptr,
			    idx_t const *restrict const ind,
			    val_t const *restrict const val,
			    val_t const *restrict const len,
			    da_iapq_t * const *const knng,
			    da_tsearch_t * restrict const ts,
			    size_t const ncands,
			    da_iakv_t * restrict const anbrs,
			    idx_t * const *restrict const cnts)
{

  ptr_t register j;
  ptr_t register n;
  idx_t register k;
  idx_t register cid;
  val_t register qsim;
  ptr_t nadd;
  val_t csim, msim, v;


  idx_t const kk = ts->k;
  idx_t const bsz = ts->bsz;
  idx_t const nthr = ts->nthreads;
  bit *restrict const cbit = ts->cbit;
  da_iakv_t *restrict const qcands = ts->qcands;
  idx_t *restrict const cnt = ts->cnt;
  ptr_t const qsz = ptr[qid + 1] - ptr[qid];
  idx_t const *const qind = ind + ptr[qid];
  val_t const *const qval = val + ptr[qid];
  val_t const *const qlen = len + ptr[qid];
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  val_t *restrict const qrow = ts->qrow;
  val_t *restrict const qrln = ts->qrln;
#else
  sdhash *restrict const qhash = ts->qhash;
#endif
  da_iapq_t *restrict const lnn = knng[qid];
  lnn->locator = ts->loc;
  nadd = 0;


  etimer_start (ts->timer_5);	/* query hash time */
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
#if PL2KNNA_PRUNE_NCACHE
  for (j = 0; j < qsz; ++j)
    {				/* cache query row */
      qrow[qind[j]] = qval[j];
      qrln[qind[j]] = qlen[j];
    }
#else
  for (j = 0; j < qsz; ++j)
    {				/* cache query row */
      qrow[qind[j]] = qval[j];
    }
#endif
#else
  sdhashSet (qhash, qind, qval, qsz);
#endif
  etimer_stop (ts->timer_5);	/* query hash time */


#if defined(Pl2KNNA_HINSERT) && Pl2KNNA_HINSERT


  etimer_start (ts->timer_6);	/* dot-product time */
  /* compute dot-products */
  da_iapqInitLocator (lnn);
  qsim = da_iapqSeeTopValSafe (lnn);
  for (n = 0, j = 0; j < ncands; ++j)
    {
      cid = qcands[j].key;
      if (ba_value (cbit, cid))
	{
	  csim = da_iapqSeeTopValSafe (knng[cid]);
	  msim = da_min (qsim, csim);
	  if (msim > PL2KNNA_PRUNE_MIN)
	    {
	      /* attempt to use pruning */
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
#if PL2KNNA_PRUNE_NCACHE
	      v =
		psdpdotp2 (msim, ptr[cid + 1] - ptr[cid], ind + ptr[cid],
			   val + ptr[cid], len + ptr[cid], qrow, qrln);
#else
	      v =
		psdpdotp (msim, ptr[cid + 1] - ptr[cid], ind + ptr[cid],
			  val + ptr[cid], qrow);
#endif
#else
	      v =
		sdhashDotp (qhash, ind + ptr[cid], val + ptr[cid],
			    ptr[cid + 1] - ptr[cid]);
#endif
	    }
	  else
	    {
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
	      v =
		psddotp (ptr[cid + 1] - ptr[cid], ind + ptr[cid],
			 val + ptr[cid], qrow);
#else
	      v =
		sdhashDotp (qhash, ind + ptr[cid], val + ptr[cid],
			    ptr[cid + 1] - ptr[cid]);
#endif
	    }
	  qcands[j].val = v;
	  n++;
	  if (v > qsim)
	    {
	      da_iapqInsert (lnn, cid, v);
	      qsim = da_iapqSeeTopValSafe (lnn);
	      nadd++;
	    }
	  else if (v < 0)
	    {
	      ts->nPruneLength++;
	    }
	  else if (v < msim)
	    {
	      qcands[j].val = -1.0;
	    }

	}
    }
  da_iapqResetLocator (lnn);
  ts->nDotProducts += n;
  etimer_stop (ts->timer_6);	/* dot-product time */

#else


  etimer_start (ts->timer_6);	/* dot-product time */
  /* compute dot-products */
  qsim = da_iapqSeeTopValSafe (lnn);
  for (n = 0, j = 0; j < ncands; ++j)
    {
      cid = qcands[j].key;
      if (ba_value (cbit, cid))
	{
	  csim = da_iapqSeeTopValSafe (knng[cid]);
	  msim = da_min (qsim, csim);
	  if (msim > PL2KNNA_PRUNE_MIN)
	    {
	      /* attempt to use pruning */
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
#if PL2KNNA_PRUNE_NCACHE
	      v =
		psdpdotp2 (msim, ptr[cid + 1] - ptr[cid], ind + ptr[cid],
			   val + ptr[cid], len + ptr[cid], qrow, qrln);
#else
	      v =
		psdpdotp (msim, ptr[cid + 1] - ptr[cid], ind + ptr[cid],
			  val + ptr[cid], qrow);
#endif
#else
	      v =
		sdhashDotp (qhash, ind + ptr[cid], val + ptr[cid],
			    ptr[cid + 1] - ptr[cid]);
#endif
	    }
	  else
	    {
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
	      v =
		psddotp (ptr[cid + 1] - ptr[cid], ind + ptr[cid],
			 val + ptr[cid], qrow);
#else
	      v =
		sdhashDotp (qhash, ind + ptr[cid], val + ptr[cid],
			    ptr[cid + 1] - ptr[cid]);
#endif
	    }
	  qcands[j].val = v;
	  n++;
	  if (v < 0)
	    {
	      ts->nPruneLength++;
	    }
	  else if (v < msim)
	    {
	      qcands[j].val = -1.0;
	    }
	}
    }
  ts->nDotProducts += n;
  etimer_stop (ts->timer_6);	/* dot-product time */

  etimer_start (ts->timer_7);	/* local update time */
  da_iapqInitLocator (lnn);
  n = ncands;
#if defined(PL2KNNA_PRESELECT) && PL2KNNA_PRESELECT
  if (ncands > kk)
    {
      n = da_iakvkselectd (ncands, kk, qcands);
    }
#endif
  j = 0;
  if (checkfull)
    {
      while (j < n && lnn->nnodes < kk)
	{
	  if (qcands[j].val > 0)
	    {
	      da_iapqInsert (lnn, qcands[j].key, qcands[j].val);
	      nadd++;
	    }
	  j++;
	}
    }
  qsim = da_iapqSeeTopVal (lnn);
  for (; j < n; ++j)
    {
      cid = qcands[j].key;
      if (qcands[j].val > qsim)
	{
	  da_iapqInsert (lnn, cid, qcands[j].val);
	  qsim = da_iapqSeeTopVal (lnn);
	  nadd++;
	}
    }
  da_iapqResetLocator (lnn);
  etimer_stop (ts->timer_7);	/* local update time */

#endif

  etimer_start (ts->timer_5);	/* query hash time */
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  for (j = 0; j < qsz; ++j)
    {
      qrow[qind[j]] = 0.0;	/* reset query row */
    }
#else
  sdhashUnset (qhash, qind, qsz);
#endif
  etimer_stop (ts->timer_5);	/* query hash time */


  /* bucket neighbors for each thread */
  etimer_start (ts->timer_8);	/* shuffle time */
  CSRMAKE (j, nthr, cnt);
  memcpy (cnts[qid - start], cnt, sizeof (idx_t) * (nthr + 1));
  for (j = 0; j < ncands; ++j)
    {
      k = (qcands[j].key / bsz);
      anbrs[cnt[k]++] = qcands[j];
    }
  etimer_stop (ts->timer_8);	/* shuffle time */

  return nadd;

}



inline ptr_t
pl2knnChooseInitialNeighbors (const da_csr_t * const docs,
			      da_tsearch_t * const ts,
			      da_iakv_t * const qcands,
			      idx_t * const cnt, const idx_t qid)
{

  ptr_t register j;
  idx_t register c1;
  idx_t register cs1;
  idx_t register ce1;
  idx_t register c2;
  idx_t register cs2;
  idx_t register ce2;
  idx_t register cid;
  val_t register cv1;
  val_t register cv2;
  ptr_t ncands;

  int const nthr = ts->nthreads;
  idx_t const bsz = ts->bsz;
  idx_t const nn = ts->nn;
  bit *restrict const cbit = ts->cbit;

  ptr_t const *restrict const rowptr = docs->rowptr;	/* index in rowind/rowval where each row starts */
  idx_t const *restrict const rowind = docs->rowind;	/* column ids in each row */
  val_t const *restrict const rowval = docs->rowval;	/* values in each row */
  ptr_t const *restrict const colptr = docs->colptr;	/* index in rowind/rowval where each row starts */
  idx_t const *restrict const colind = docs->colind;	/* column ids in each col */
  val_t const *restrict const colval = docs->colval;	/* values in each col */
  memset (cnt, 0, nthr * sizeof (idx_t));	/* zero out per-thread counts */

  j = rowptr[qid];
  ptr_t const lj = rowptr[qid + 1];
  c1 = rowind[j];
  cv1 = rowval[j];
  cs1 = colptr[c1];
  ce1 = colptr[c1 + 1];
  cv2 = cs2 = ce2 = 0;
  if (j + 1 < lj)
    {
      j++;
      c2 = rowind[j];
      cv2 = rowval[j];
      cs2 = colptr[c2];
      ce2 = colptr[c2 + 1];
    }
  else
    {
      c2 = -1;
    }
  ncands = 0;
  while (ncands < nn)
    {
      if (c2 < 0)
	{			/* last col has been reached. */
	  for (; cs1 < ce1 && ncands < nn; ++cs1)
	    {
	      if (colind[cs1] == qid)
		{
		  continue;
		}
	      if (ba_value (cbit, colind[cs1]))
		{		/* bit set means candidate not yet chosen */
		  cid = colind[cs1];
		  qcands[ncands++] = (da_iakv_t)
		  {
		  cid, 0.0};
		  cnt[(cid / bsz)]++;	/* track count of each thread's assignment */
		  ba_toggle (cbit, cid);
		}
	    }
	  break;
	}
      else if (c1 < 0)
	{			/* last col has been reached. */
	  for (; cs2 < ce2 && ncands < nn; ++cs2)
	    {
	      if (colind[cs2] == qid)
		{
		  continue;
		}
	      if (ba_value (cbit, colind[cs2]))
		{		/* bit set means candidate not yet chosen */
		  cid = colind[cs2];
		  qcands[ncands++] = (da_iakv_t)
		  {
		  cid, 0.0};
		  cnt[(cid / bsz)]++;	/* track count of each thread's assignment */
		  ba_toggle (cbit, cid);
		}
	    }
	  break;
	}

      if (colval[cs1] * cv1 > colval[cs2] * cv2)
	{
	  if (colind[cs1] != qid)
	    {
	      if (ba_value (cbit, colind[cs1]))
		{		/* bit set means candidate not yet chosen */
		  cid = colind[cs1];
		  qcands[ncands++] = (da_iakv_t)
		  {
		  cid, 0.0};
		  cnt[(cid / bsz)]++;	/* track count of each thread's assignment */
		  ba_toggle (cbit, cid);
		}
	    }
	  cs1++;

	}
      else
	{
	  if (colind[cs2] != qid)
	    {
	      if (ba_value (cbit, colind[cs2]))
		{		/* bit set means candidate not yet chosen */
		  cid = colind[cs2];
		  qcands[ncands++] = (da_iakv_t)
		  {
		  cid, 0.0};
		  cnt[(cid / bsz)]++;	/* track count of each thread's assignment */
		  ba_toggle (cbit, cid);
		}
	    }
	  cs2++;
	}

      if (cs1 == ce1)
	{			/* end of col 1 - switch to another col, if exists */
	  if (j + 1 < lj)
	    {
	      j++;
	      c1 = rowind[j];
	      cv1 = rowval[j];
	      cs1 = colptr[c1];
	      ce1 = colptr[c1 + 1];
	    }
	  else
	    c1 = -1;
	}
      else if (cs2 == ce2)
	{			/* end of col 2 - switch to another col, if exists */
	  if (j + 1 < lj)
	    {
	      j++;
	      c2 = rowind[j];
	      cv2 = rowval[j];
	      cs2 = colptr[c2];
	      ce2 = colptr[c2 + 1];
	    }
	  else
	    c2 = -1;
	}
    }

  /* clear bit vector */
  for (j = 0; j < ncands; ++j)
    {
      ba_toggle (cbit, qcands[j].key);
    }

  return ncands;
}



size_t
pl2knnEnhanceNeighborhood (params_t * const params,
			   da_csr_t * restrict const docs,
			   da_csr_t * restrict const fdocs,
			   da_csr_t * restrict const bnbrs,
			   idx_t * restrict const nnbrs,
			   val_t const *const rlen,
			   da_iapq_t * const *const knng,
			   da_iakv_t * const *const qnn,
			   idx_t * const *const cnts,
			   da_tsearch_t * const *const search)
{

  idx_t register i;
  ptr_t register j;
  ptr_t register nadd;
  idx_t progressInd, pct, nqrows;
  ptr_t nadded;

  idx_t const nrows = docs->nrows;	/* num rows */
  idx_t const kk = params->k;
  idx_t const nn = kk * params->alpha;	/* number of neighbors' neighbors to retrieve */
  int const nthr = params->nthreads;
  idx_t const bsz = ceil (nrows / nthr);

  for (nadded = 0, i = 0; i < nthr; ++i)
    {
      nadded -= search[i]->nHeapInserts;
    }

  etimer_start (params->timer_6);	/* sort time */
  da_csr_t const *restrict const nbrs =
    pl2knnCreateEnhanceIndex (nn, nrows, knng, search);
  etimer_stop (params->timer_6);	/* sort time */

  if (params->verbosity > 0)
    {
      printf ("Enhance neighborhood... ");
    }
  da_progress_init (pct, progressInd, nrows);

  etimer_start (params->timer_7);	/* actual search time */
  for (i = 0; i < nrows; i += params->nqrows)
    {
      nqrows = da_min (params->nqrows, nrows - i);

	/** current block neighbors */
      etimer_start (params->timer_8);	/* neighborhood extraction time */
      knng2csr2 (bnbrs, knng + i, nqrows);
      etimer_stop (params->timer_8);	/* neighborhood extraction time */

#pragma omp parallel default(none) shared(nqrows, i)
      {

	idx_t register ii;
	ptr_t register j;
	ptr_t register k;
	ptr_t nadd, ncands;
	idx_t qid, cid;
	da_iakv_t *anbrs;

	int const tid = omp_get_thread_num ();
	da_tsearch_t *restrict const ts = search[tid];
	bit *restrict const cbit = ts->cbit;
	da_iakv_t *restrict const qcands = ts->qcands;
	idx_t *restrict const cnt = ts->cnt;
	ptr_t const *restrict const ptr = docs->rowptr;
	idx_t const *restrict const ind = docs->rowind;
	val_t const *restrict const val = docs->rowval;
	ptr_t const *restrict const rptr = nbrs->rowptr;
	idx_t const *restrict const rind = nbrs->rowind;
	val_t const *restrict const rval = nbrs->rowval;
	ptr_t const *restrict const cptr = nbrs->colptr;
	idx_t const *restrict const cind = nbrs->colind;
	val_t const *restrict const cval = nbrs->colval;
	ptr_t const *restrict const bptr = bnbrs->rowptr;
	idx_t const *restrict const bind = bnbrs->rowind;
#if PL2KNNA_PRUNE_NCACHE
	ptr_t const *restrict const fptr = fdocs->rowptr;
	idx_t const *restrict const find = fdocs->rowind;
	val_t const *restrict const fval = fdocs->rowval;
#endif

	etimer_start (ts->timer_1);	/* parallel time */

	etimer_start (ts->timer_2);	/* tile search time */
#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC)
	for (ii = 0; ii < nqrows; ii++)
	  {
	    qid = i + ii;

	    if (nnbrs[qid] < kk)
	      {
		continue;	/* already found all the neighbors */
	      }

	    anbrs = qnn[ii];

	    etimer_start (ts->timer_4);	/* candidate selection time */
	    /* zero out per-thread counts */
	    memset (cnt, 0, nthr * sizeof (idx_t));

	    /* tag current neighbors */
	    for (j = rptr[qid]; j < rptr[qid + 1]; ++j)
	      {
		ba_assign_false (cbit, rind[j]);
	      }
	    for (j = bptr[ii]; j < bptr[ii + 1]; ++j)
	      {
		ba_assign_false (cbit, bind[j]);
	      }

	    /* find neighbor's neighbors */
	    for (ncands = 0, j = rptr[qid]; j < rptr[qid + 1]; ++j)
	      {
		cid = rind[j];

		for (k = rptr[cid]; k < rptr[cid + 1]; ++k)
		  {		/* neighbors's neighbors */
#ifdef PRUNENGBS
		    if (ncands >= nn || rval[k] < rval[j])
		      {
			break;
		      }
#else
		    if (ncands >= nn)
		      {
			break;
		      }
#endif

		    if (rind[k] != qid && ba_value (cbit, rind[k]))
		      {
			cnt[(rind[k] / bsz)]++;
			qcands[ncands++] = (da_iakv_t)
			{
			rind[k], 0.0};
			ba_assign_false (cbit, rind[k]);
		      }
		  }
	      }
	    etimer_stop (ts->timer_4);	/* candidate selection time */

	    etimer_start (ts->timer_3);	/* tag time */
	    for (j = 0; j < ncands; ++j)
	      {
		ba_assign_true (cbit, qcands[j].key);
	      }
	    etimer_stop (ts->timer_3);	/* tag time */

	    if (ncands == 0)
	      {
		/* resent remote counts */
		CSRMAKE (j, nthr, cnt);
		memcpy (cnts[ii], cnt, sizeof (idx_t) * (nthr + 1));
		/* untag current neighbors */
		for (j = rptr[qid]; j < rptr[qid + 1]; ++j)
		  {
		    ba_assign_true (cbit, rind[j]);
		  }
		for (j = bptr[ii]; j < bptr[ii + 1]; ++j)
		  {
		    ba_assign_true (cbit, bind[j]);
		  }
		ASSERT (ba_count (cbit, nrows) == nrows);
		continue;
	      }
	    ts->nCandidates += ncands;

#if defined(PL2KNNA_PRUNE) && PL2KNNA_PRUNE
#if PL2KNNA_PRUNE_NCACHE
	    ts->nHeapInserts +=
	      pl2knnCheckCandidatesPrune (qid, i, false, fptr, find, fval,
					  rlen, knng, ts, ncands, anbrs,
					  cnts);
#else
	    ts->nHeapInserts +=
	      pl2knnCheckCandidatesPrune (qid, i, false, ptr, ind, val, knng,
					  ts, ncands, anbrs, cnts);
#endif
#else
	    ts->nHeapInserts +=
	      pl2knnCheckCandidates (qid, i, false, ptr, ind, val, knng[qid],
				     ts, ncands, anbrs, cnts);
#endif

	    /* untag current neighbors */
	    etimer_start (ts->timer_3);	/* tag time */
	    for (j = rptr[qid]; j < rptr[qid + 1]; ++j)
	      {
		ba_assign_true (cbit, rind[j]);
	      }
	    for (j = bptr[ii]; j < bptr[ii + 1]; ++j)
	      {
		ba_assign_true (cbit, bind[j]);
	      }
	    etimer_stop (ts->timer_3);	/* tag time */

	    ASSERT (ba_count (cbit, nrows) == nrows);

	  }
	etimer_stop (ts->timer_2);	/* tile search time */

#pragma omp barrier

	/** go through the lists and update neighbors assigned to the thread */
	etimer_start (ts->timer_9);	/* candidate update time */
	for (nadd = 0, j = 0; j < nqrows; j++)
	  {
	    qid = i + j;
	    anbrs = qnn[j];
	    for (k = cnts[j][tid]; k < cnts[j][tid + 1]; ++k)
	      {
		cid = anbrs[k].key;
		if (anbrs[k].val > da_iapqSeeTopVal (knng[cid]))
		  {
		    if (da_iapqInsertHeap (knng[cid], qid, anbrs[k].val))
		      {
			nadd++;
		      }
		  }
	      }
	  }
	ts->nHeapInserts += nadd;
	etimer_stop (ts->timer_9);	/* candidate update time */
	etimer_stop (ts->timer_1);	/* parallel time */

      }

      if (params->verbosity > 0 && i % progressInd == 0)
	da_progress_advance (pct);
    }
  etimer_stop (params->timer_7);	/* actual search time */
  if (params->verbosity > 0){
    da_progress_finalize (pct);
    printf ("\n");
  }

  da_csr_Free ((da_csr_t **) & nbrs);

  for (i = 0; i < nthr; ++i)
    {
      nadded += search[i]->nHeapInserts;
    }
  return nadded;
}





void
pl2knnFindInitialNeighbors (params_t * const params,
			    da_csr_t * const docs,
			    da_csr_t * const fdocs,
			    da_csr_t * const bnbrs,
			    idx_t * const nnbrs,
			    val_t * const rlen,
			    da_iapq_t * const *const knng,
			    da_iakv_t * const *const qnn,
			    idx_t * const *const cnts,
			    da_tsearch_t * const *const search)
{


  idx_t register i;
  idx_t nqrows, progressInd, pct;

  idx_t const nrows = docs->nrows;	// num rows
  idx_t const nn = params->k * params->alpha;	/* number of neighbor's neighbors to retrieve */

  if (params->alpha < 1)
    {
      da_errexit
	("-alpha must be at least 1. Random selection not implemented for this mode.");
    }

  /* build inverted index and sort both rows and columns in decreasing value oder */
  etimer_start (params->timer_6);	/* sort time */
  da_csr_pCreateIndex_a (docs, DA_COL, true);
  da_pl2knnSortTopKValues (docs, DA_ROW, nn, 0, search);
  da_pl2knnSortTopKValues (docs, DA_COL, nn, 0, search);
#if PL2KNNA_PRUNE_NCACHE
  /* pre-compute suffix l2-norms for fdocs */
#pragma omp parallel default(none)
  {
    ptr_t register j;
    val_t v;
    ptr_t const *restrict const ptr = fdocs->rowptr;
    idx_t const *restrict const ind = fdocs->rowind;
    val_t const *restrict const val = fdocs->rowval;

#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC_FAST)
    for (i = 0; i < nrows; ++i)
      {
	for (v = 1.0, j = ptr[i]; j < ptr[i + 1]; ++j)
	  {
	    v -= val[j] * val[j];
	    rlen[j] = sqrt (v);
	  }
      }
  }

#endif
  etimer_stop (params->timer_6);	/* sort time */

  if (params->verbosity > 0)
    {
      printf ("nn: %d\n", nn);
      printf ("Find initial neighbors... ");
    }

  etimer_start (params->timer_7);	/* actual search time */
  da_progress_init (pct, progressInd, nrows);
  for (i = 0; i < nrows; i += params->nqrows)
    {
      nqrows = da_min (params->nqrows, nrows - i);

      etimer_start (params->timer_8);	/* neighborhood extraction time */
      knng2csr2 (bnbrs, knng + i, nqrows);
      etimer_stop (params->timer_8);	/* neighborhood extraction time */

#pragma omp parallel default(none) shared(nqrows, i)
      {

	ptr_t register j;
	ptr_t register k;
	idx_t qid, cid;
	ptr_t ncands, nadd;
	da_iakv_t *anbrs;

	int const tid = omp_get_thread_num ();
	da_tsearch_t *restrict const ts = search[tid];
	bit *restrict const cbit = ts->cbit;
	da_iakv_t *restrict const qcands = ts->qcands;
	idx_t *restrict const cnt = ts->cnt;
	ptr_t const *restrict const ptr = docs->rowptr;
	idx_t const *restrict const ind = docs->rowind;
	val_t const *restrict const val = docs->rowval;
	ptr_t const *restrict const bptr = bnbrs->rowptr;
	idx_t const *restrict const bind = bnbrs->rowind;
#if PL2KNNA_PRUNE_NCACHE
	ptr_t const *restrict const fptr = fdocs->rowptr;
	idx_t const *restrict const find = fdocs->rowind;
	val_t const *restrict const fval = fdocs->rowval;
#endif

	etimer_start (ts->timer_1);	/* parallel time */

	etimer_start (ts->timer_2);	/* tile search time */
#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC)
	for (j = 0; j < nqrows; j++)
	  {
	    qid = i + j;
	    if (ptr[qid + 1] == ptr[qid])
	      {
		continue;	/* empty row, no neighbors */
	      }

	    etimer_start (ts->timer_4);	/* candidate selection time */
	    ncands =
	      pl2knnChooseInitialNeighbors (docs, ts, qcands, cnt, qid);
	    if (ncands == 0)
	      {
		etimer_stop (ts->timer_4);	/* candidate selection time */
		continue;
	      }
	    ts->nCandidates += ncands;
	    etimer_stop (ts->timer_4);	/* candidate selection time */

	    /* mark existing neighbors */
	    etimer_start (ts->timer_3);	/* tag time */
	    for (k = bptr[j]; k < bptr[j + 1]; ++k)
	      {
		ba_assign_false (cbit, bind[k]);
	      }
	    etimer_stop (ts->timer_3);	/* tag time */
#if defined(PL2KNNA_PRUNE) && PL2KNNA_PRUNE
#if PL2KNNA_PRUNE_NCACHE
	    ts->nHeapInserts +=
	      pl2knnCheckCandidatesPrune (qid, i, true, fptr, find, fval,
					  rlen, knng, ts, ncands, qnn[j],
					  cnts);
#else
	    ts->nHeapInserts +=
	      pl2knnCheckCandidatesPrune (qid, i, true, ptr, ind, val, knng,
					  ts, ncands, qnn[j], cnts);
#endif
#else
	    ts->nHeapInserts +=
	      pl2knnCheckCandidates (qid, i, true, ptr, ind, val, knng[qid],
				     ts, ncands, qnn[j], cnts);
#endif
	    /* un-mark existing neighbors */
	    etimer_start (ts->timer_3);	/* tag time */
	    for (k = bptr[j]; k < bptr[j + 1]; ++k)
	      {
		ba_assign_true (cbit, bind[k]);
	      }
	    etimer_stop (ts->timer_3);	/* tag time */

	  }
	etimer_stop (ts->timer_2);	/* tile search time */

#pragma omp barrier

	/** go through the lists and update neighbors assigned to the thread */

	etimer_start (ts->timer_9);	/* candidate update time */
	for (nadd = 0, j = 0; j < nqrows; j++)
	  {
	    qid = i + j;
	    anbrs = qnn[j];
	    for (k = cnts[j][tid]; k < cnts[j][tid + 1]; ++k)
	      {
		cid = anbrs[k].key;
		if (anbrs[k].val > da_iapqSeeTopValSafe (knng[cid]))
		  {
		    da_iapqInsertHeap (knng[cid], qid, anbrs[k].val);
		    nadd++;
		  }
	      }
	  }
	ts->nHeapInserts += nadd;
	etimer_stop (ts->timer_9);	/* candidate update time */
	etimer_stop (ts->timer_1);	/* parallel time */

      }

      if (params->verbosity > 0 && i % progressInd == 0)
	da_progress_advance (pct);
    }
  if (params->verbosity > 0){
    da_progress_finalize (pct);
    printf ("\n");
  }

  /* update neighbor counts */
  for (i = 0; i < nrows; i++)
    {
      nnbrs[i] = knng[i]->nnodes;
    }
  etimer_stop (params->timer_7);	/* actual search time */

}


/**
 * Reorder the document matrix in decreasing number of column nnzs.
 */
void
pl2knnReorderCols (da_csr_t * const docs, char const order)
{

  da_iikv_t *const orderColNnz =
    da_iikvnmalloc (docs->ncols, "pl2knnReorderCols: orderColNnz");
  idx_t *const cperm = da_inmalloc (docs->ncols, "pl2knnReorderCols: cperm");
  int const ntr = omp_get_max_threads ();
  ptr_t **const sizes =
    da_nmalloc_a (ntr * sizeof (ptr_t *), "pl2knnReorderCols: sizes");


#pragma omp parallel default(none)
  {

    ptr_t register i;
    ptr_t register j;

    int const tid = omp_get_thread_num ();
    idx_t const ncols = docs->ncols;
    ptr_t const nnz = docs->rowptr[docs->nrows];
    ptr_t const bsz = ceil (nnz / (float) ntr);
    ptr_t const cbsz = ceil (ncols / (float) ntr);
    ptr_t const *const ptr = docs->rowptr;
    idx_t *const ind = docs->rowind;
    ptr_t *const sz = sizes[tid] =
      da_pnmalloc_a (docs->ncols, "pl2knnReorderCols: sz");

	/** execute local count */
    for (i = tid * bsz; i < da_min (nnz, (tid + 1) * bsz); ++i)
      {
	sz[ind[i]]++;
      }

#pragma omp barrier

	/** aggregate counts */
    for (j = 1; j < ntr; ++j)
      {
	for (i = tid * cbsz; i < da_min (ncols, (tid + 1) * cbsz); ++i)
	  {
	    sizes[0][i] += sizes[j][i];
	  }
      }

#pragma omp barrier

	/** sort columns by count */
    for (i = tid * cbsz; i < da_min (ncols, (tid + 1) * cbsz); ++i)
      {
	orderColNnz[i].key = i;
	orderColNnz[i].val = sizes[0][i];
      }

#pragma omp barrier

#pragma omp single
    {
      if (order == 'i')
	{
	  da_iikvsorti (ncols, orderColNnz);
	}
      else
	{
	  da_iikvsortd (ncols, orderColNnz);
	}
    }
    for (i = tid * cbsz; i < da_min (ncols, (tid + 1) * cbsz); ++i)
      {
	cperm[orderColNnz[i].key] = i;
      }

#pragma omp barrier

	/** renumber columns */
    for (i = tid * bsz; i < da_min (nnz, (tid + 1) * bsz); ++i)
      {
	ind[i] = cperm[ind[i]];
      }

  }

  da_csr_SortIndices (docs, DA_ROW);

  for (int i = 0; i < ntr; ++i)
    {
      da_free ((void **) &sizes[i], LTERM);
    }
  da_free ((void **) &orderColNnz, &sizes, &cperm, LTERM);

}
