/*!
 \file  pl2knn.c
 \brief This file contains pL2Knn related functions

 \author David C. Anastasiu
 */

#include "includes.h"
#include "sdhash.h"
#include "pl2knn.h"



inline void
init_tsearch (size_t const tid,
	      idx_t const nrows,
	      idx_t const ncols,
	      idx_t const mrl,
	      params_t const *const params, da_tsearch_t * const ts)
{
  ptr_t i;
  ts->alpha = params->alpha;
  ts->beta = params->beta;
  ts->gamma = params->gamma;
  ts->delta = params->delta;
  ts->k = params->k;
  ts->nn = params->k * params->alpha;
  ts->nbl = params->nbl;
  ts->nqrows = params->nqrows;
  ts->ndrows = params->ndrows;
  ts->ninnzs = params->ninnzs;
  ts->nthreads = params->nthreads;
  ts->bsz = ceil (nrows / params->nthreads);
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  ts->qrow =
    (val_t *) da_nmalloc_a (ncols * sizeof (val_t), "init_tsearch: ts->qrow");
  ts->qrln =
    (val_t *) da_nmalloc_a (ncols * sizeof (val_t), "init_tsearch: ts->qrln");
#else
  ts->qhash = sdhashCreate (mrl, PL2KNN_QHASH_HTSIZE, ncols);
#endif
  ts->cbit = ba_new (nrows);
  ba_complement (ts->cbit, nrows);
  ts->qcands =
    (da_iakv_t *) da_nmalloc_a (da_max (ts->k, ts->nn) * sizeof (da_iakv_t),
				"init_tsearch: qcands");
  ts->scands =
    (da_ivkv_t *) da_nmalloc_a (da_max (nrows, ncols) * sizeof (da_ivkv_t),
				"init_tsearch: scands");
  ts->cnt =
    (idx_t *) da_nmalloc_a ((params->nthreads + 1) * sizeof (idx_t),
			    "init_search: cnt");
  ts->loc = (ptr_t *) da_malloc (nrows * sizeof (ptr_t), "init_tsearch: loc");
  for (i = 0; i < nrows; ++i)
    {
      ts->loc[i] = -1;
    }
}

void
free_tsearch (da_tsearch_t ** tsp)
{
  size_t i;

  da_tsearch_t *ts = *tsp;
#if PL2KNN_QHASH == PL2KNN_QHASH_HASH
  sdhashFree (&ts->qhash);
#endif
  da_free ((void **) &ts->cnt, &ts->loc, &ts->qcands, &ts->scands, &ts->cbit,
	   &ts->qrow, &ts->qrln, &ts->lcands, &ts->dcands, &ts->accum, tsp,
	   LTERM);
}

void
free_all_tsearch (da_tsearch_t *** searchp, idx_t const sz)
{
  size_t i;
  da_tsearch_t **search = *searchp;

  for (i = 0; i < sz; ++i)
    {
      free_tsearch (&(search[i]));
    }
  da_free ((void **) searchp, LTERM);
}



/**
 * Main entry point to L2KNNG.
 */
void
pl2knnFindNeighbors (params_t * const params)
{

  idx_t register i;
  idx_t register j;
  idx_t pct = 0;
  idx_t mrl, arl;
  ptr_t cnt;
  val_t simt;
  val_t *rlen = NULL;

  if (params->mode == MODE_PL2KNN && params->nqrows > params->ndrows)
  {
    if(params->verbosity > -1)
      printf("  Note: Adjusting nqrows to %d. Cannot be greater than ndrows.\n",
	 params->ndrows);
    params->nqrows = params->ndrows;
  }
  if(params->verbosity > -1)
    printf ("ndrows: %d, nqrows: %d, ninnzs: %zu\n", params->ndrows,
	  params->nqrows, params->ninnzs);

  /* pre-process data */
  preProcessData (params);
  da_csr_CompactColumns (params->docs);
  da_csr_SortIndices (params->docs, DA_ROW);

  /* make input aligned */
  da_csr_t *const docs = da_csr_Copy_a (params->docs);
  da_csr_Free (&params->docs);
  params->docs = docs;
  da_csr_t *const vdocs = da_csr_Copy_a (docs);	/* value sorted docs */

  omp_set_num_threads (params->nthreads);

  // allocate memory
  int const nthr = params->nthreads;
  idx_t const nrows = docs->nrows;	// num rows
  idx_t const ncols = docs->ncols;	// num cols
  ptr_t const nnz = docs->rowptr[nrows];
  val_t *const rval = docs->rowval;
  idx_t *const rind = docs->rowind;
  ptr_t *const rptr = docs->rowptr;
  idx_t const kk = params->k;
  idx_t const nn = kk * params->alpha;
  idx_t nqrows = params->nqrows;
  idx_t const ndrows = params->ndrows;
  idx_t const ninnzs = params->ninnzs;
  da_iakv_t **qnn =
    da_iakvnmatalloc_a (nqrows, da_max (kk, kk * params->alpha),
			"pl2knnFindNeighbors: qnn");
  idx_t **cnts = da_inmatalloc_a (nqrows, (nthr + 1), "pl2knnFindNeighbors: rlocs");	/* shared locator for all the priority queues, one for each thread */
  idx_t *const nnbrs = da_inmalloc_a (nrows, "pl2knnFindNeighbors: nnbrs");	/* number of potential neighbors <= k */
  da_csr_t *bnbrs = da_csr_Create ();
  da_csr_Alloc (bnbrs, nqrows, nrows, nn * nqrows, DA_ROW, true);
  da_iapq_t **knng = params->knng =
    (da_iapq_t **) da_malloc (nrows * sizeof (da_iapq_t *),
			      "knng");	/* KNNG result as set of priority queues */
  for (mrl = 0, arl = 0, i = 0; i < nrows; ++i)
    {
      knng[i] = da_iapqCreateNoLocator (params->k, nrows);
      if (rptr[i + 1] - rptr[i] > mrl)
	{
	  mrl = rptr[i + 1] - rptr[i];	/* find max row length */
	}
      arl += rptr[i + 1] - rptr[i];
    }
  da_tsearch_t **search =
    (da_tsearch_t **) da_malloc (params->nthreads * sizeof (da_tsearch_t *),
				 "pl2knnFindNeighbors: search");
  ba_init ();
#pragma omp parallel default(none) shared(mrl, search)
  {
    int const tid = omp_get_thread_num ();
    search[tid] =
      (da_tsearch_t *) da_nmalloc (sizeof (da_tsearch_t),
				   "pl2knnFindNeighbors: search[i]");
    init_tsearch (tid, nrows, ncols, mrl, params, search[tid]);
  }
#if PL2KNNA_PRUNE_NCACHE
  rlen = da_vmalloc (nnz, "pl2knnFindNeighbors: rlen");	/* row suffix lengths */
#endif


  arl /= (double) nrows;
  params->mrl = mrl;
  if(params->verbosity > -1){
    da_csr_PrintInfo (docs, "input matrix:", "\n");
    printf ("maxrl %d, meanrl %d\n", mrl, arl);
    pl2knnPrintCompileOptions ();
#if PL2KNN_QHASH == PL2KNN_QHASH_HASH
    printf ("\tqhash htsize: %zu, maxsz: %zu\n", search[0]->qhash->htsize,
  	    search[0]->qhash->maxsz);
    fflush (stdout);
#endif
  }

    /*** STEP1: Initial KNNG selection based on large values + find rows with less than k neighbors. ***/
  timer_start (params->timer_9);	// initial neighbors time
  pl2knnFindInitialNeighbors (params, vdocs, docs, bnbrs, nnbrs, rlen, knng,
			      qnn, cnts, search);
  timer_stop (params->timer_9);	// initial neighbors time
  if(params->verbosity > -1)
    da_printTimerLong ("\tInitial knng build time: ", params->timer_9);
  params->timer_3 += params->timer_9;
  timer_clear (params->timer_9);
#ifdef EXTRATIMES
  if(params->verbosity > -1){
    da_printTimerLong ("\tIndex and sort time: ", params->timer_6);
    timer_clear (params->timer_6);
    da_printTimerLong ("\tActual search: ", params->timer_7);
    timer_clear (params->timer_7);
    da_printThreadTimers ("Thread parallel section times", 1, search);
    da_printThreadTimers ("Thread search times", 2, search);
    da_printThreadTimers ("Thread candidate tag times", 3, search);
    da_printThreadTimers ("Thread candidate selection times", 4, search);
    da_printThreadTimers ("Thread query hash times", 5, search);
    da_printThreadTimers ("Thread dot-product times", 6, search);
    da_printThreadTimers ("Thread shuffle times", 8, search);
    da_printThreadTimers ("Thread local update times", 7, search);
    da_printThreadTimers ("Thread candidate update times", 9, search);
    for (i = 0; i < params->nthreads; ++i)
    {
      timer_clear (search[i]->timer_1);
      timer_clear (search[i]->timer_2);
      timer_clear (search[i]->timer_3);
      timer_clear (search[i]->timer_4);
      timer_clear (search[i]->timer_5);
      timer_clear (search[i]->timer_6);
      timer_clear (search[i]->timer_7);
      timer_clear (search[i]->timer_8);
      timer_clear (search[i]->timer_9);
    }
    printf ("\n");
  }
#endif

  /* verify approximate results */
  if (params->vFile)
    verify_knng_results2 (knng, nrows, params->vFile, 1);

    /*** STEP2: Enhance initial neighbors with neighbor's neighbors. ***/
  if (params->gamma)
    {
      timer_start (params->timer_9);	// enhance neighbors time
      for (i = 0; i < params->gamma; ++i)
	{
	  cnt =
	    pl2knnEnhanceNeighborhood (params, vdocs, docs, bnbrs, nnbrs,
				       rlen, knng, qnn, cnts, search);
	  if(params->verbosity > -1) 
            printf ("# improvements: %zu\n", cnt);
	  if (cnt < params->delta * nrows * params->k)
	    break;
	}
      timer_stop (params->timer_9);	// enhance neighbors time
      if(params->verbosity > -1)
         da_printTimerLong ("\tEnhance initial knng time: ", params->timer_9);
      params->timer_3 += params->timer_9;
      timer_clear (params->timer_9);

#ifdef EXTRATIMES
      if(params->verbosity > -1){
        da_printTimerLong ("\tIndex and sort time: ", params->timer_6);
        timer_clear (params->timer_6);
        da_printTimerLong ("\tActual search: ", params->timer_7);
        timer_clear (params->timer_7);
        da_printThreadTimers ("Thread parallel section times", 1, search);
        da_printThreadTimers ("Thread search times", 2, search);
        da_printThreadTimers ("Thread candidate tag times", 3, search);
        da_printThreadTimers ("Thread candidate selection times", 4, search);
        da_printThreadTimers ("Thread query hash times", 5, search);
        da_printThreadTimers ("Thread dot-product times", 6, search);
        da_printThreadTimers ("Thread shuffle times", 8, search);
        da_printThreadTimers ("Thread local update times", 7, search);
        da_printThreadTimers ("Thread candidate update times", 9, search);
        for (i = 0; i < params->nthreads; ++i)
  	{
	  timer_clear (search[i]->timer_1);
	  timer_clear (search[i]->timer_2);
	  timer_clear (search[i]->timer_3);
	  timer_clear (search[i]->timer_4);
	  timer_clear (search[i]->timer_5);
	  timer_clear (search[i]->timer_6);
	  timer_clear (search[i]->timer_7);
	  timer_clear (search[i]->timer_8);
	  timer_clear (search[i]->timer_9);
	}
        printf ("\n");
      }
#endif

      /* verify approximate results */
      if (params->vFile)
	verify_knng_results2 (knng, nrows, params->vFile, 1);

    }

  if(params->verbosity > -1){
    pl2knnTransferCounts (params, search);
    printf ("Num. heap inserts: %zu\n", params->nHeapInserts);
#if defined(PL2KNNA_PRUNE) && PL2KNNA_PRUNE
    printf ("Num. approximate candidates pruned by l2-norm bound: %zu\n",
  	    params->nPruneLength);
    params->nDotProducts -= params->nPruneLength;
#endif
  }

  if (params->mode == MODE_PL2KNN_A)
    {				/* only interested in the approximate solution */

      /* report number of similarity pairs found */
      for (i = 0; i < nrows; ++i)
	{
	  params->nSimPairs += knng[i]->nnodes;
	}

      /* save neighbor graph */
      da_save_knng (params, knng, nrows, DA_FMT_CLUTO);

      /* free memory */
      da_csr_FreeAll (&bnbrs, &vdocs, LTERM);
      for (i = 0; i < nrows; ++i)
	{
	  knng[i]->locator = NULL;
	  da_iapqDestroy (knng[i]);
	}
      da_imatfree (&cnts, params->nqrows, params->nthreads);
      da_iakvmatfree (&qnn, params->nqrows,
		      params->k + params->k * params->alpha);
      free_all_tsearch (&search, params->nthreads);
      da_free ((void **) &knng, &rlen, &nnbrs, LTERM);

      return;
    }

    /*** STEP4: Verify found neighbors, adjusting lists as appropriate. ***/

  timer_start (params->timer_9);
  if (params->verbosity > 0)
    {
      printf ("Verifying ");
      fflush (stdout);
    }

  /* memory allocation time */
  timer_start (params->timer_5);
  ptr_t mnnz = da_min (nnz, ndrows * mrl);
  da_csr_t *const idx = da_csr_Create ();	/* forward and inverted index of the DB tile */
  da_csr_Alloc (idx, ndrows, ncols, mnnz, DA_ROW, true);
  da_csr_Alloc (idx, ndrows, ncols, mnnz, DA_COL, true);
  idx->rowids = da_imalloc_a (ndrows, "pl2knnFindNeighbors: idx->rowids");	/* original ids for permuted rows */
  idx->colids = da_ismalloc_a (nrows, -1, "pl2knnFindNeighbors: idx->colids");	/* inverse permutation of rows */
  idx->rnorms = da_vmalloc_a (mnnz, "pl2knnFindNeighbors: idx->rnorms");	/* suffix norms for forward index */
  idx->cnorms = da_vmalloc_a (mnnz, "pl2knnFindNeighbors: idx->cnorms");	/* suffix norms for inverted index */
  for (i = 0; i < nqrows; ++i)
    {				/* make sure query candidate lists have enough space */
      qnn[i] =
	da_iakvrealloc_a (qnn[i], ndrows,
			  "pl2knnFindNeighbors realloc: qnn[i]");
    }
  da_iakv_t *const msim = da_iakvmalloc (nrows, "pl2knnFindNeighbors: msim");	/* minimum neighborhood similarities */
  ptr_t *const pszs = da_pnmalloc (ndrows + 1, "pl2knnFindNeighbors: rend");	/* prefix sizes -- counts/pointer for end of inverted index */

  /* update thread search structures */
#pragma omp parallel default(none) shared(search)
  {
    int const tid = omp_get_thread_num ();
    da_tsearch_t *const ts = search[tid];
    ts->dcands =
      da_inmalloc_a (ts->ndrows, "pl2knnFindNeighbors: ts->dcands");
    ts->lcands =
      da_inmalloc_a (ts->ndrows, "pl2knnFindNeighbors: ts->lcands");
    ts->accum =
      da_vsmalloc_a (ts->ndrows, L2KNN_CINIT,
		     "pl2knnFindNeighbors: ts->accum");
    da_free ((void **) &ts->qcands, &ts->cbit, LTERM);
  }
  timer_stop (params->timer_5);

  /* permute matrix columns */
  etimer_start (params->timer_6);	/* sorting time */
  pl2knnReorderCols (docs, 'i');

  /* compute and store suffix lengths for each object */
  if (!rlen)
    {
      rlen = da_vmalloc (nnz, "pl2knnFindNeighbors: rlen");
    }
  docs->rnorms = rlen;
#pragma omp parallel default(none)
  {
    ptr_t register j;
    val_t v;
    ptr_t const *const ptr = docs->rowptr;
    idx_t const *const ind = docs->rowind;
    val_t const *const val = docs->rowval;
    val_t *const len = docs->rnorms;

#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC_FAST)
    for (i = 0; i < nrows; ++i)
      {
	for (v = 1.0, j = ptr[i]; j < ptr[i + 1]; ++j)
	  {
	    v -= val[j] * val[j];
	    len[j] = v > 0 ? sqrt (v) : 0.0;
	  }
	msim[i].key = i;
      }
  }
  etimer_stop (params->timer_6);	/* sorting time */


  /* execute search */

  idx_t lqid, start, end, n, k, qid, cid, nidx;
  ptr_t nadd;
  da_iakv_t *anbrs;
  val_t rct;
  start = nidx = 0;
  bnbrs->ncols = nrows;
  while (start < nrows)
    {
	/** create index for the next search block -- index starts with row b as 0 */
      etimer_start (params->timer_4);	/* indexing time */
      n =
	pl2knnIndexP (nrows - start, msim + start, params, knng, docs, pszs,
		      idx);
      end = start + n;
      nidx++;
      if (idx->colptr[idx->ncols] == 0)
	{
	  start = end;
	  continue;
	}
	/** add the current inverted index size **/
      params->indexSize += idx->colptr[idx->ncols];
      etimer_stop (params->timer_4);	/* indexing time */

      /* tile the search for overlapping indexed rows */
      rct = msim[start].val;
      for (lqid = start; lqid < end; lqid += params->nqrows)
	{
	  nqrows = da_min (params->nqrows, end - lqid);

	    /** current block neighbors */
	  etimer_start (params->timer_8);	/* neighborhood extraction time */
	  knng2csr2s (bnbrs, knng + lqid, nqrows, start, end);	/* filter out items before start and after end */
	  etimer_stop (params->timer_8);	/* neighborhood extraction time */

#pragma omp parallel default(none) shared(search, bnbrs, qnn, cnts, knng, nqrows, rct, lqid, start, n)
	  {
	    int const tid = omp_get_thread_num ();
	    idx_t j, k, qid, cid;
	    ptr_t nadd;
	    da_iakv_t *anbrs;

	    da_tsearch_t *const ts = search[tid];
	    ts->bsz = ceil (n / (float) nthr);
	    val_t *const accum = ts->accum;

	    /* find neighbors */
#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC)
	    for (i = 0; i < nqrows; ++i)
	      {
		pl2KnnFindQueryNeighbors (ts, msim[lqid + i].key,
					  lqid - start + i, i, rct, docs, idx,
					  bnbrs, knng, qnn[i], cnts[i],
					  msim + start);
	      }

#pragma omp barrier

	    /* update candidate neighborhoods */
	    etimer_start (ts->timer_9);	/* candidate update time */
	    for (nadd = 0, j = 0; j < nqrows; j++)
	      {
		qid = msim[lqid + j].key;
		anbrs = qnn[j];
		for (k = cnts[j][tid]; k < cnts[j][tid + 1]; ++k)
		  {
		    cid = anbrs[k].key;
		    if (anbrs[k].val > da_iapqSeeTopVal (knng[cid]))
		      {
			da_iapqInsertHeap (knng[cid], qid, anbrs[k].val);
			nadd++;
		      }
		  }
	      }
	    ts->nHeapInserts += nadd;
	    etimer_stop (ts->timer_9);	/* candidate update time */

	  }
	}

      /* now tile the search for remaining rows */
      for (lqid = end; lqid < nrows; lqid += params->nqrows)
	{
	  nqrows = da_min (params->nqrows, nrows - lqid);

#pragma omp parallel default(none) shared(search, bnbrs, qnn, cnts, knng, nqrows, rct, lqid, start, n)
	  {
	    int const tid = omp_get_thread_num ();
	    idx_t j, k, qid, cid;
	    ptr_t nadd;
	    da_iakv_t *anbrs;

	    da_tsearch_t *const ts = search[tid];

	    /* find neighbors */
#pragma omp for schedule(dynamic, PL2KNN_DYNAMIC)
	    for (i = 0; i < nqrows; ++i)
	      {
		pl2KnnFindQueryNeighbors (ts, msim[lqid + i].key, -1, -1, rct,
					  docs, idx, bnbrs, knng, qnn[i],
					  cnts[i], msim + start);
	      }

#pragma omp barrier

	    /* update candidate neighborhoods */
	    etimer_start (ts->timer_9);	/* candidate update time */
	    for (nadd = 0, j = 0; j < nqrows; j++)
	      {
		qid = msim[lqid + j].key;
		anbrs = qnn[j];
		for (k = cnts[j][tid]; k < cnts[j][tid + 1]; ++k)
		  {
		    cid = anbrs[k].key;
		    if (anbrs[k].val > da_iapqSeeTopVal (knng[cid]))
		      {
			da_iapqInsertHeap (knng[cid], qid, anbrs[k].val);
			nadd++;
		      }
		  }
	      }
	    ts->nHeapInserts += nadd;
	    etimer_stop (ts->timer_9);	/* candidate update time */

	  }
	}


      /* mark processed objects in inverse permutation */
      for (lqid = start; lqid < end; ++lqid)
	{
	  idx->colids[msim[lqid].key] = -1;
	}

      if (params->verbosity > 0)
	{
	  while (pct < 100 * start / (float) nrows)
	    {
	      da_progress_advance (pct);
	    }
	}

      start = end;

    }
  timer_stop (params->timer_9);
  params->timer_3 += params->timer_9;

  pl2knnTransferCounts (params, search);

  /* report number of similarity pairs found */
  for (params->nSimPairs = 0, i = 0; i < nrows; ++i)
    {
      params->nSimPairs += knng[i]->nnodes;
    }

  if (params->verbosity > 0)
    {
      da_progress_finalize (pct);
      printf ("\n");
      da_printTimerLong ("\tVerification time: ", params->timer_9);
      printf ("Total num. heap inserts: %zu\n", params->nHeapInserts);
      printf ("Number of db tiles: %d\n", nidx);
#ifndef EXTRACOUNTS
      printf ("Dynamic index size: %zu\n", params->indexSize);
#endif


#ifdef EXTRATIMES
      if(params->verbosity > -1){
        da_printThreadTimers ("Thread candidate generation times", 1, search);
        da_printThreadTimers ("Thread candidate verification times", 2, search);
        da_printThreadTimers ("Thread candidate tag times", 3, search);
        da_printThreadTimers ("Thread shuffle times", 8, search);
        da_printThreadTimers ("Thread candidate update times", 9, search);
        printf ("\n");
      }
#endif
      double cg, cv;
      for (cg = 0, cv = 0, i = 0; i < params->nthreads; ++i)
	{
	  if (search[i]->timer_1 > cg)
	    {
	      cg = search[i]->timer_1;
	    }
	  if (search[i]->timer_2 > cv)
	    {
	      cv = search[i]->timer_2;
	    }
	}
      params->timer_1 = cg;
      params->timer_2 = cv;

    }

  /* verify results */
  if (params->vFile)
    verify_knng_results2 (knng, nrows, params->vFile, params->verbosity);

  /* save neighbor graph */
  da_save_knng (params, knng, nrows, params->fmtWrite);

  /* free memory */
  da_csr_FreeAll (&bnbrs, &vdocs, &idx, LTERM);
  for (i = 0; i < nrows; ++i)
    {
      knng[i]->locator = NULL;
      da_iapqDestroy (knng[i]);
    }
  da_imatfree (&cnts, params->nqrows, (params->nthreads + 1));
  da_iakvmatfree (&qnn, params->nqrows,
		  params->k + params->k * params->alpha);
  free_all_tsearch (&search, params->nthreads);
  da_free ((void **) &knng, &nnbrs, &msim, &pszs, LTERM);

}


/**
 * Find neighbors when query is some row in the DB tile
 */
void
pl2KnnFindQueryNeighbors (da_tsearch_t * const ts, idx_t const rid,	/* global query row id in original doc order */
			  idx_t const dqid,	/* local id in db tile order -- row id in idx */
			  idx_t const lqid,	/* local id in query tile order -- row id in bnbrs */
			  val_t const rct,	/* minimum neighborhood similarity among indexed objects */
			  da_csr_t const *const docs,	/* input documents */
			  da_csr_t const *const idx,	/* db tile index */
			  da_csr_t const *const bnbrs,	/* current version of tile neighborhoods */
			  da_iapq_t * const *const knng,	/* the output graph */
			  da_iakv_t * const qnn,	/* candidate update list for query row */
			  idx_t * const qcnt,	/* counts array for query row */
			  da_iakv_t * const msim)	/* minimum neighborhood similarities and current processing order */
{

  ptr_t register i;
  ptr_t register j;
  ptr_t register n;
  idx_t register k;
  idx_t lcid, ncands, nskip;
  ptr_t dps, nadd;
  val_t bnd, qms, qcms, v, acc;

  idx_t const bsz = ts->bsz;
  idx_t const nthr = ts->nthreads;
  idx_t *const dcands = ts->dcands;
  idx_t *const lcands = ts->lcands;
  accum_t *const accum = ts->accum;
  idx_t *const cnt = ts->cnt;
  ptr_t const *const iptr = docs->rowptr;
  idx_t const *const iind = docs->rowind;
  val_t const *const ival = docs->rowval;
  val_t const *const ilen = docs->rnorms;
  ptr_t const *const rptr = idx->rowptr;
  idx_t const *const rind = idx->rowind;
  val_t const *const rval = idx->rowval;
  val_t const *const rlen = idx->rnorms;
  idx_t const *const rids = idx->rowids;
  idx_t const *const rinv = idx->colids;
  ptr_t const *const cptr = idx->colptr;
  idx_t const *const cind = idx->colind;
  val_t const *const cval = idx->colval;
  val_t const *const clen = idx->cnorms;
  ptr_t const *const nrptr = bnbrs->rowptr;
  idx_t const *const nrind = bnbrs->rowind;
  val_t const *const nrval = bnbrs->rowval;
#if PL2KNN_QHASH == PL2KNN_QHASH_DENSE
  val_t *const qrow = ts->qrow;
  val_t *const qrln = ts->qrln;
#else
  sdhash *const qhash = ts->qhash;
#endif
  da_iapq_t *const lnn = knng[rid];
  da_iakv_t *const qheap = lnn->heap;
  qms = qheap[0].val;		/* query vector minimum neighborhood similarity */

  timer_start (ts->timer_1);	/* candidate generation time */

  /* initialize local heap locator */
  lnn->locator = ts->loc;
  da_iapqInitLocator (lnn);

  /* mark items that we already have answers for:
   *      current neighborhood, reverse neighborhood */
  etimer_start (ts->timer_3);	/* tag time */
  nskip = 0;
  if (dqid >= 0)
    {
      accum[dqid] = L2KNN_CPRUNED;
    }
  if (lqid >= 0)
    {
      for (i = nrptr[lqid]; i < nrptr[lqid + 1]; ++i)
	{
	  lcid = rinv[nrind[i]];	/* candidate id in the realm of tile ids */
	  if (lcid >= 0 && accum[lcid] == L2KNN_CINIT)
	    {
	      lcands[nskip++] = lcid;
	      accum[lcid] = L2KNN_CPRUNED_GRAPH;
	    }
	}
    }
  for (i = 0; i < ts->k; ++i)
    {
      lcid = rinv[qheap[i].key];	/* candidate id in the realm of tile ids */
      if (lcid >= 0 && accum[lcid] == L2KNN_CINIT)
	{
	  lcands[nskip++] = lcid;
	  accum[lcid] = L2KNN_CPRUNED_NBRS;
	}
    }
  etimer_stop (ts->timer_3);	/* tag time */

    /*** find matches for query row ***/
  ncands = 0;
  for (bnd = 1.0, ncands = 0, i = iptr[rid]; i < iptr[rid + 1]; ++i)
    {
      v = ival[i];
      k = iind[i];

      for (j = cptr[k]; j < cptr[k + 1]; ++j)
	{
	  lcid = cind[j];	/* potential candidate */

	  if (accum[lcid] > 0.0)
	    {
	      // accumulate
	      accum[lcid] += v * cval[j];

#ifdef L2CG
	      // check l2-norm bound
	      if (accum[lcid] + bnd * clen[j] < da_min (qms, msim[lcid].val))
		{
		  accum[lcid] = L2KNN_CPRUNED;
#ifdef EXTRACOUNTS
		  ts->nPruneLength++;
#endif
		}
#endif
	    }
	  else if (bnd >= rct && accum[lcid] > L2KNN_CPRUNED)
	    {
	      // accumulate
	      if (accum[lcid] == L2KNN_CINIT)
		{
		  accum[lcid] = v * cval[j];
		  dcands[ncands++] = lcid;
		}
	      else
		{
		  accum[lcid] += v * cval[j];
		}

#ifdef L2CG
	      // check l2-norm bound
	      if (accum[lcid] + bnd * clen[j] < da_min (qms, msim[lcid].val))
		{
		  accum[lcid] = L2KNN_CPRUNED;
#ifdef EXTRACOUNTS
		  ts->nPruneLength++;
#endif
		}
#endif
	    }

	}

      bnd = ilen[i];
      /* cache the query row */
      qrow[k] = v;
      qrln[k] = ilen[i];
    }

  timer_stop (ts->timer_1);	/* candidate generation time */
  timer_start (ts->timer_2);	/* candidate verification time */

    /*** process candidates ***/
  ts->nCandidates += ncands;
  for (dps = 0, nadd = 0, i = 0; i < ncands; i++)
    {
      lcid = dcands[i];

      if (accum[lcid] <= 0.0)
	{
	  accum[lcid] = L2KNN_CINIT;
	  continue;
	}

      acc = accum[lcid];
      accum[lcid] = L2KNN_CINIT;
      j = rptr[lcid];
#if defined(PSCV) || defined(L2CV)
      qcms = da_min (qms, msim[lcid].val);	/* min sim of either the query or candidate neighborhoods */
#endif

#ifdef PSCV
      if (j < rptr[lcid + 1] && (acc + rlen[j] < qcms) && qrow[rind[j]] == 0)
	{
#ifdef EXTRACOUNTS
	  ts->nPrunePscore++;
#endif
	  continue;
	}
#endif

      /* compute the remaining portion of the dot product */
      for (; j < rptr[lcid + 1]; ++j)
	{
	  if (qrow[rind[j]] > 0)
	    {
	      acc += rval[j] * qrow[rind[j]];

	      /* check L2-norm bound */
#ifdef L2CV
	      if (acc + rlen[j] * qrln[rind[j]] < qcms)
		{
#ifdef EXTRACOUNTS
		  ts->nPruneLength2++;
#endif
		  goto nextcand;
		}
#endif
	    }
	}

      dps++;

      if (acc > qms)
	{
	  da_iapqInsert (lnn, rids[lcid], acc);
	  qms = qheap[0].val;
	  nadd++;
	}
      if (acc > qcms)
	{
	  accum[lcid] = acc;
	}
      else
	{
	  accum[lcid] = L2KNN_CINIT;
	}

    nextcand:
      continue;
    }


  ts->nHeapInserts += nadd;
  ts->nDotProducts += dps;

  /* clear the accumulation array for the pre-computed candidates */
  etimer_start (ts->timer_3);	/* tag time */
  for (i = 0; i < nskip; ++i)
    {
      accum[lcands[i]] = L2KNN_CINIT;
    }
  if (dqid >= 0)
    {
      accum[dqid] = L2KNN_CINIT;
    }
  etimer_stop (ts->timer_3);	/* tag time */


    /** set up candidate neighborhood transfer and clear remaining items in accum array */
  etimer_start (ts->timer_8);	/* shuffle time */
  memset (cnt, 0, nthr * sizeof (idx_t));
  for (i = 0; i < ncands; i++)
    {
      lcid = dcands[i];
      if (accum[lcid] > 0 && msim[lcid].val < accum[lcid])
	{
	  cnt[(lcid / bsz)]++;
	}
      else
	{
	  accum[lcid] = L2KNN_CINIT;
	}
    }
  CSRMAKE (j, nthr, cnt);
  memcpy (qcnt, cnt, sizeof (idx_t) * (nthr + 1));
  for (i = 0; i < ncands; i++)
    {
      if (accum[dcands[i]] > 0)
	{
	  k = (dcands[i] / bsz);
	  qnn[cnt[k]].key = rids[dcands[i]];
	  qnn[cnt[k]++].val = accum[dcands[i]];
	  accum[dcands[i]] = L2KNN_CINIT;
	}
    }
  etimer_stop (ts->timer_8);	/* shuffle time */

  /* reset the heap locator for the current row */
  da_iapqResetLocator (lnn);

  /* un-cache the query row */
  for (j = iptr[rid]; j < iptr[rid + 1]; ++j)
    {
      qrow[iind[j]] = 0;
    }

  timer_stop (ts->timer_2);	/* candidate verification time */

}



idx_t
pl2knnIndexP (idx_t const nsims,
	      da_iakv_t * const msim,
	      params_t * const params,
	      da_iapq_t * const *const knng,
	      da_csr_t const *const docs,
	      ptr_t * const pszs, da_csr_t * const idx)
{


  idx_t register i;
  ptr_t register j;
  idx_t rid, n;
  val_t qms, isim;
  ptr_t b, e, l, nnz;

  idx_t const ncols = docs->ncols;
  idx_t const ndrows = da_min (params->ndrows, nsims);
  ptr_t const ninnz = params->ninnzs;

  /* find new order */
  for (i = 0; i < nsims; ++i)
    {
      msim[i].val = da_iapqSeeTopVal (knng[msim[i].key]);
    }
  da_iakvsorti (nsims, msim);
  isim = msim[0].val;

  /* create inverted index for tile */
  int const nt = omp_get_max_threads ();
  da_csr_t **tmps =
    (da_csr_t **) da_malloc (sizeof (da_csr_t *) * nt,
			     "da_csr_pCreateIndex_a: tmps");

#pragma omp parallel default(none) shared(tmps, n, isim)
  {
    idx_t register i;
    ptr_t register j;
    idx_t rid;
    ptr_t cs, k, b, l, nnz;
    ptr_t *ptr;
    idx_t *ind;
    val_t *val, *len;
    da_csr_t *tmp = NULL;

    int const tid = omp_get_thread_num ();

    ptr_t const *const iptr = docs->rowptr;
    idx_t const *const iind = docs->rowind;
    val_t const *const ival = docs->rowval;
    val_t const *const ilen = docs->rnorms;
    ptr_t *const rptr = idx->rowptr;
    idx_t *const rind = idx->rowind;
    val_t *const rval = idx->rowval;
    val_t *const rlen = idx->rnorms;
    idx_t *const rids = idx->rowids;
    ptr_t *const cptr = idx->colptr;
    idx_t *const cind = idx->colind;
    val_t *const cval = idx->colval;
    val_t *const clen = idx->cnorms;
    idx_t *const rinv = idx->colids;	/* Note: using colids for inverse permutation of row ids */

    /* figure out prefix/suffix lengths */
    idx_t rbsz = ceil (ndrows / (float) nt);
    idx_t s = tid * rbsz;
    idx_t e = da_min (ndrows, (tid + 1) * rbsz);
    for (i = s; i < e; ++i)
      {
	rid = msim[i].key;
	k = pszs[i] = iptr[rid + 1] - iptr[rid];
	for (j = iptr[rid]; j < iptr[rid + 1]; ++j)
	  {
		/** check where to stop indexing **/
	    if (ilen[j] < isim)
	      {
		pszs[i] = j - iptr[rid] + 1;
		break;
	      }
	  }
	rptr[i] = k - pszs[i];
      }

#pragma omp barrier
#pragma omp single
    {
      /* figure out num rows to index */
      for (j = 0, i = 0; i < ndrows; ++i)
	{
	  j += pszs[i];

	  if (j >= ninnz)
	    {
	      break;
	    }
	}
      if (i == 0)
	{
	  ++i;
	}
      idx->nrows = i;
      CSRMAKE (j, i, rptr);
    }

    n = idx->nrows;

    /* create the temporary assigned tile of the input matrix */
    tmp = tmps[tid] = da_csr_Create ();
    idx_t const cbsz = ceil (ncols / (float) nt);
    rbsz = ceil (n / (float) nt);
    s = tid * rbsz;
    e = da_min (n, (tid + 1) * rbsz);
    tmp->nrows = e - s;
    tmp->ncols = ncols;

    /* create forward index and inv. index pointer */
    for (i = s; i < e; ++i)
      {
	rid = msim[i].key;
	b = iptr[rid] + pszs[i];
	l = iptr[rid + 1] - b;
	if (rptr[i + 1] - rptr[i] != l)
	  {
	    da_errexit ("ERR: rptr[i+1]-rptr[i] != l, %zu != %zu",
			rptr[i + 1] - rptr[i], l);
	  }
	if (l == 0)
	  {
	    continue;
	  }
	/* copy row data */
	memcpy (rind + rptr[i], iind + b, sizeof (idx_t) * l);
	memcpy (rval + rptr[i], ival + b, sizeof (val_t) * l);
	memcpy (rlen + rptr[i], ilen + b, sizeof (val_t) * l);
      }

    /* figure out nnz for chunk we're working with and allocate inversion space */
    for (nnz = 0, i = s; i < e; ++i)
      {
	rid = msim[i].key;
	nnz += iptr[rid + 1] - iptr[rid];
      }
    ptr = tmp->colptr = da_pnmalloc (ncols + 1, "pl2knnIndexP: ptr");
    ind = tmp->colind = da_imalloc (nnz, "pl2knnIndexP: ind");
    val = tmp->colval = da_vmalloc (nnz, "pl2knnIndexP: val");
    len = tmp->cnorms = da_vmalloc (nnz, "pl2knnIndexP: len");

    /* transpose assigned chunk of the matrix */
    for (i = s; i < e; ++i)
      {
	rid = msim[i].key;
	rids[i] = rid;		/* update row id of permuted row */
	rinv[rid] = i;		/* inverse permutation */
	for (j = iptr[rid]; j < iptr[rid] + pszs[i]; ++j)
	  {
	    ptr[iind[j]]++;
	  }
      }
    CSRMAKE (i, ncols, ptr);
    for (i = s; i < e; ++i)
      {
	rid = msim[i].key;
	for (j = iptr[rid]; j < iptr[rid] + pszs[i]; ++j)
	  {
	    b = ptr[iind[j]];
	    ind[b] = i;
	    val[b] = ival[j];
	    len[b] = ilen[j];
	    ptr[iind[j]]++;
	  }
      }
    CSRSHIFT (i, ncols, ptr);


#pragma omp barrier

    /* create reverse pointer array */
    for (j = (tid * cbsz); j < da_min ((tid + 1) * cbsz, ncols); ++j)
      {
	for (k = 0, i = 0; i < nt; ++i)
	  {
	    k += tmps[i]->colptr[j + 1] - tmps[i]->colptr[j];
	  }
	cptr[j] = k;
      }

#pragma omp barrier
#pragma omp single
    {
      CSRMAKE (i, ncols, cptr);
    }

    /* copy thread indexes into the master inverted index */
    for (j = (tid * cbsz); j < da_min ((tid + 1) * cbsz, ncols); ++j)
      {
	for (i = 0; i < nt; ++i)
	  {
	    cs = tmps[i]->colptr[j];
	    k = tmps[i]->colptr[j + 1] - cs;
	    da_icopy (k, tmps[i]->colind + cs, cind + cptr[j]);
	    da_vcopy (k, tmps[i]->colval + cs, cval + cptr[j]);
	    da_vcopy (k, tmps[i]->cnorms + cs, clen + cptr[j]);
	    cptr[j] += k;
	  }
      }

#pragma omp barrier

    /* free tmp memory */
    tmps[tid] = NULL;
    da_csr_Free (&tmp);
  }
  CSRSHIFT (i, ncols, idx->colptr);	/* shift col pointer */

  da_free ((void **) &tmps, LTERM);

  return idx->nrows;
}





/**
 * Find similar rows in the matrix -  this version of the function reports
 * the number of candidates/dot products that were considered in the search.
 * \param mat The CSR matrix we're searching in
 * \param rid Row we're looking for neighbors for
 * \param noSelfSim Do not return self-similarity as one of the results
 * \param nqterms length of query sparse vector
 * \param qind Column indices of the query sparse vector
 * \param qval Values in the query sparse vector
 * \param simtype Which similarity to compute (DA_SIM_COS or DA_SIM_JAC)
 * \param nsim Number of similar pairs to get (-1 to get all)
 * \param hits Array or length mat->nrows to hold values for possible matches and result
 * \param i_marker Optional marker array of length mat->nrows to mark candidates
 * \param i_cand Optional key-value array of length mat->nrows to store and sort candidates
 * \param ncands Reference to int variable to hold number of candidates
 *
 * \return Number of similar pairs found
 */
idx_t
da_GetKSimilarRows (da_csr_t * mat, idx_t rid, char noSelfSim,
		    idx_t nqterms, idx_t * qind, val_t * qval, char simtype,
		    idx_t nsim, da_ivkv_t * hits, idx_t * i_marker,
		    da_ivkv_t * i_cand, idx_t * ncands)
{
  ssize_t i, ii, j, k;
  idx_t nrows, ncols, ncand;
  ptr_t *colptr;
  idx_t *colind, *marker;
  val_t *colval, *rnorms, mynorm, *rsums, mysum;
  da_ivkv_t *cand;

  if (nqterms == 0)
    return 0;

  nrows = mat->nrows;
  ncols = mat->ncols;
  colptr = mat->colptr;
  colind = mat->colind;
  colval = mat->colval;

  marker =
    (i_marker ? i_marker :
     da_ismalloc (nrows, -1, "da_csr_GetSimilarSmallerRows: marker"));
  cand =
    (i_cand ? i_cand :
     da_ivkvmalloc (nrows, "da_csr_GetSimilarSmallerRows: cand"));

  switch (simtype)
    {
    case DA_SIM_COS:
      for (ncand = 0, ii = 0; ii < nqterms; ii++)
	{
	  i = qind[ii];
	  if (i < ncols)
	    {
	      for (j = colptr[i]; j < colptr[i + 1]; j++)
		{
		  k = colind[j];
		  if (noSelfSim && k == rid)
		    continue;
		  if (marker[k] == -1)
		    {
		      cand[ncand].key = k;
		      cand[ncand].val = 0;
		      marker[k] = ncand++;
		    }
		  cand[marker[k]].val += colval[j] * qval[ii];
		}
	    }
	}
      break;

    case DA_SIM_JAC:
      for (ncand = 0, ii = 0; ii < nqterms; ii++)
        {
          i = qind[ii];
          if (i < ncols)
            {
              for (j = colptr[i]; j < colptr[i + 1]; j++)
                {
                  k = colind[j];
                  if (noSelfSim && k == rid)
                    continue;
                  if (marker[k] == -1)
                    {
                      cand[ncand].key = k;
                      cand[ncand].val = 0;
                      marker[k] = ncand++;
                    }
                  cand[marker[k]].val += colval[j] * qval[ii];
                }
            }
        }

      rnorms = mat->rnorms;
      mynorm = da_vdot (nqterms, qval, 1, qval, 1);

      for (i = 0; i < ncand; i++)
        cand[i].val =
          cand[i].val / (rnorms[cand[i].key] + mynorm - cand[i].val);
      break;


    default:
      da_errexit ("Unknown similarity measure %d\n", simtype);
      return 0;
    }

  *ncands = ncand;

  /* clear markers */
  for (j = 0, i = 0; i < ncand; i++)
    marker[cand[i].key] = -1;

  if (nsim == -1 || nsim >= ncand)
    {
      nsim = ncand;
    }
  else
    {
      nsim = da_min (nsim, ncand);
      da_ivkvkselectd (ncand, nsim, cand);
    }
  da_ivkvcopy (nsim, cand, hits);

  if (i_marker == NULL)
    da_free ((void **) &marker, LTERM);
  if (i_cand == NULL)
    da_free ((void **) &cand, LTERM);

  return nsim;
}



/**
 * Reorder the document matrix in decreasing number of column nnzs.
 * Also compute and store necessary stats about the data.
 * Store initial order in cperm.
 */
void
l2knnReorderCols (da_csr_t * docs, idx_t * csizes, idx_t * cperm)
{
  ssize_t i, j, k, nnz;
  idx_t nrows, ncols;
  da_iikv_t *orderColNnz = NULL;
  da_ivkv_t *orderColVal = NULL;
  char pv = 0;
  ptr_t *rowptr;
  idx_t *rowind, *tsizes = NULL;
  val_t v, *rowval, *cval = NULL;

  nrows = docs->nrows;
  ncols = docs->ncols;
  rowptr = docs->rowptr;
  rowind = docs->rowind;
  nnz = rowptr[nrows];
#if COLSORTBY == COLSORTBY_MAX || COLSORTBY == COLSORTBY_AVG
  rowval = docs->rowval;
  cval = da_vsmalloc (ncols, 0.0, "l2knnReorderCols: cval");
#endif

  //record col nnzs
  da_iset (ncols, 0, csizes);
  for (i = 0; i < nnz; i++)
    {
#if COLSORTBY == COLSORTBY_MAX
      j = rowind[i];
      if (rowval[j] > cval[j])
        cval[j] = rowval[j];
      csizes[j]++;
#elif COLSORTBY == COLSORTBY_AVG
      j = rowind[i];
      cval[j] += rowval[j];
      csizes[j]++;
#else
      csizes[rowind[i]]++;
#endif
    }

#if COLSORTBY == COLSORTBY_AVG
  for (i = 0; i < ncols; ++i)
    if (csizes[i] > 0)
      cval[i] /= csizes[i];
#endif

#if COLSORTBY == COLSORTBY_MAX || COLSORTBY == COLSORTBY_AVG
  orderColVal = da_ivkvmalloc (ncols, "l2knnReorderCols: orderColNnz");
  //get new column order
  for (j = 0; j < ncols; j++)
    {
      orderColVal[j].key = j;
      orderColVal[j].val = cval[j];
    }
  da_ivkvsorti (ncols, orderColVal);    // sort columns

  for (j = 0; j < ncols; j++)
    {
      cperm[orderColVal[j].key] = j;    // store column permutation
      csizes[j] = orderColVal[j].val;   // new column size after permuting
    }
#elif COLSORTBY == COLSORTBY_SIZE
  orderColNnz = da_iikvmalloc (ncols, "l2knnReorderCols: orderColNnz");
  //get new column order
  for (j = 0; j < ncols; j++)
    {
      orderColNnz[j].key = j;
      orderColNnz[j].val = csizes[j];
    }
  da_iikvsorti (ncols, orderColNnz);    // sort columns

  for (j = 0; j < ncols; j++)
    {
      cperm[orderColNnz[j].key] = j;    // store column permutation
      csizes[j] = orderColNnz[j].val;   // new column size after permuting
    }
#elif COLSORTBY == COLSORTBY_RAND
  da_irandArrayPermute (ncols, cperm, 10 * ncols, 1);
  tsizes = da_imalloc (ncols, "l2knnReorderCols: tsizes");
  for (j = 0; j < ncols; j++)
    {
      tsizes[j] = csizes[cperm[j]];     // new column size after permuting
    }
  da_icopy (ncols, tsizes, csizes);
#else
  da_errexit ("Invalid COLSORTBY criteria in defs.h.");
#endif

  /* set new column ids in docs */
  for (i = 0; i < nnz; i++)
    rowind[i] = cperm[rowind[i]];

  da_csr_SortIndices (docs, DA_ROW);

  da_free ((void **) &orderColNnz, &orderColVal, &cval, &tsizes, LTERM);

}






/**
 * Reorder the document matrix in decreasing number of column nnzs.
 * Also compute and store necessary stats about the data.
 * Store initial order in cperm.
 */
void
l2apReorderCols (da_csr_t * docs, idx_t * csizes, idx_t * cperm)
{
  ssize_t i, j, k, nnz;
  idx_t nrows, ncols;
  da_iikv_t *orderColNnz = NULL;
  char pv = 0;
  ptr_t *rowptr;
  idx_t *rowind;
  val_t v, *rowval;

  nrows = docs->nrows;
  ncols = docs->ncols;
  rowptr = docs->rowptr;
  rowind = docs->rowind;
  nnz = rowptr[nrows];

  //record col nnzs
  da_iset (ncols, 0, csizes);
  for (i = 0; i < nnz; i++)
    csizes[rowind[i]]++;

  orderColNnz = da_iikvmalloc (ncols, "l2knnReorderCols: orderColNnz");
  //get new column order
  for (j = 0; j < ncols; j++)
    {
      orderColNnz[j].key = j;
      orderColNnz[j].val = csizes[j];
    }
  da_iikvsortd (ncols, orderColNnz);    // sort columns

  for (j = 0; j < ncols; j++)
    {
      if (cperm)
        cperm[orderColNnz[j].key] = j;  // store column permutation
      csizes[j] = orderColNnz[j].val;   // new column size after permuting
    }

  /* set new column ids in docs */
  for (i = 0; i < nnz; i++)
    rowind[i] = cperm[rowind[i]];

  da_csr_SortIndices (docs, DA_ROW);

  da_free ((void **) &orderColNnz, LTERM);

}
