/*!
\file  bkheap.c
\brief This file implements a sparse dot-product hash table.
*/
#include "sdhash.h"


/**
 * Create an sdhash table. Set htsize to 0 to automatically determine based on maxfid and maxnnz.
 * \param maxsz Total number of elements that will ever be added to the hash at the same time
 * \param htsize The size of the hash table (must be a power of 2)
 * \param maxfid The max size of an id that will be looked up in the table (used to estimate htsize)
 * NOTE: UNSAFE, no bounds checking for maxsz. Ensure at most maxsz elements will be set into the hash at a time.
 */
sdhash *
sdhashCreate (size_t const maxsz, size_t const htsize, size_t const maxfid)
{
  size_t h;
  sdhash *t;

  t = (sdhash *) da_nmalloc (sizeof (sdhash), "sdhashCreate: t");

  t->maxsz = maxsz;

  if (maxsz < 1)
    {
      da_errexit
	("sdhashCreate: What do you need this table for? The maxsz parameter must be non-negative. %zu is not.",
	 maxsz);
    }

  /* figure out htsize */
  if (htsize)
    {
      if (!da_ispow2 (htsize))
	{
	  da_errexit
	    ("sdhashCreate: The htsize parameter must be a power of two. %zu is not.",
	     htsize);
	}
      t->htsize = htsize;
    }
  else
    {
      if (maxfid < 1)
	{
	  da_errexit
	    ("sdhashCreate: The maxfid parameter must be non-negative in order to determine a proper htsize to use. %zu is not.",
	     maxsz);
	}
      h = da_min (maxsz * SDHASH_LF, maxfid / 2);
      if (!da_ispow2 (h))
	{
	  h = da_nextpow2 (h);
	}
      t->htsize = h;
    }
  t->htmask = t->htsize - 1;

  t->data = da_ivkvsmalloc (t->maxsz - 1, (da_ivkv_t)
			    {
			    -1, 0.0}
			    , "sdhashCreate: t->data");
  t->ind = da_ismalloc (t->htsize, IDX_MAX, "sdhashCreate: t->ind");
  t->val = da_vnmalloc (t->htsize, "sdhashCreate: t->val");

  return t;
}

/**
 * Free an sdhash structure
 */
void
sdhashFree (sdhash ** const pt)
{
  if (!pt)
    {
      return;
    }
  sdhash *const t = *pt;
  da_free ((void **) &t->data, &t->ind, &t->val, LTERM);
  da_free ((void **) pt, LTERM);
}


/**
 * Print the current contents of an sdhash structure
 */
void
sdhashPrint (sdhash const *const t)
{
  size_t register i;
  size_t register j;

  printf ("sdhash");
  printf
    ("  t->maxsz %zu\n  t->sz %zu\n  t->htsize %zu\n  t->htmask %zu\n  t->loc count: \n",
     t->maxsz, t->sz, t->htsize, t->htmask);
  for (j = 0, i = 0; i < t->htsize; ++i)
    {
      if (t->ind[i] != IDX_MAX)
	{
	  j++;
	}
    }
  printf ("\nt->ind/val nnzs (%zu): \n", j);
  for (i = 0; i < t->htsize; ++i)
    {
      if (t->ind[i] != IDX_MAX)
	{
	  printf (" (%d %f) ", t->ind[i], t->val[i]);
	}
    }

  printf ("\nt->data (%zu): \n", t->sz);
  for (i = 0; i < t->sz; ++i)
    {
      printf (" %d_(%d %f) ", t->ind[sdhashHash (t->data[i].key, t->htmask)],
	      t->data[i].key, t->data[i].val);
    }

  printf ("\n\n");
  fflush (stdout);

}


/**
 *  Find location in the hash for a given key
 */
//inline uint sdhashHash(
//        idx_t const key,
//        size_t const htmask)
//{
//    uint h = key+1;
//    // Secondary hashing discovered by Doug Lea
//    h += (h << 9);
//    h ^= (h >> 14);
//    h += (h << 4);
//    h ^= (h >> 10);
//
//    return h & htmask;
//}



/**
 * Set the query values in the hash table
 * NOTE: UNSAFE, no bounds checking for maxsz. Ensure at most maxsz elements will be set into the hash at a time.
 */
inline void
sdhashSet (sdhash * const t,
	   idx_t const *const qind, val_t const *const qval, ptr_t const qsz)
{
  uint register i;
  size_t h;
  char s;

  for (i = 0; i < qsz; ++i)
    {
      h = sdhashHash (qind[i], t->htmask);
      ASSERT (h < t->htsize);
      if (t->ind[h] == IDX_MAX)
	{
	  t->ind[h] = qind[i];
	  t->val[h] = qval[i];
	}
      else
	{			/* collision */
	  ASSERT (t->sz < t->maxsz - 1);
	  t->data[t->sz].key = qind[i];
	  t->data[t->sz++].val = qval[i];
	}
    }

  /* check for index sorted overflow */
  if (t->sz < 2)
    {
      return;
    }
  for (s = 0, i = 1; i < t->sz; ++i)
    {
      if (t->data[i - 1].key > t->data[i].key)
	{
	  s = 1;
	  break;
	}
    }
  if (s)
    {
      da_ivkvsortki (t->sz, t->data);
    }

}


/**
 * Set the query values in the hash table
 * NOTE: UNSAFE, no bounds checking for maxsz. Ensure at most maxsz elements will be set into the hash at a time.
 * NOTE2: Prefer sdhashSet over this method, as it has to sort overflow every time an item is added to overflow.
 */
inline void
sdhashSetKeyVal (sdhash * const t, idx_t const key, val_t const val)
{
  uint register i;
  size_t const h = sdhashHash (key, t->htmask);
  char s;

  if (t->ind[h] == IDX_MAX)
    {
      t->ind[h] = key;
      t->val[h] = val;
    }
  else
    {
      ASSERT (t->sz < t->maxsz - 1);
      t->data[t->sz].key = key;
      t->data[t->sz++].val = val;

      /* check for index sorted overflow */
      if (t->sz < 2)
	{
	  return;
	}
      for (s = 0, i = 1; i < t->sz; ++i)
	{
	  if (t->data[i - 1].key > t->data[i].key)
	    {
	      s = 1;
	      break;
	    }
	}
      if (s)
	{
	  da_ivkvsortki (t->sz, t->data);
	}

    }

}


/**
 * Remove the query values from the hash table
 */
inline void
sdhashUnset (sdhash * const t, idx_t const *const qind, ptr_t const qsz)
{
  size_t register i;

  for (i = 0; i < qsz; ++i)
    {
      t->ind[sdhashHash (qind[i], t->htmask)] = IDX_MAX;
    }
  t->sz = 0;
}


/**
 * Compute the dot-product between the sparse vector stored in the hash table and another sparse vector
 */
inline accum_t
sdhashDotp (sdhash * const t,
	    idx_t const *const ind, val_t const *const val, ptr_t const sz)
{
  size_t register i;
  size_t register h;
  ssize_t register f;
  ssize_t register l;
  ssize_t register m;
  idx_t id;
  accum_t register sim = 0.0;

  for (i = 0; i < sz; ++i)
    {
      h = sdhashHash (ind[i], t->htmask);
      ASSERT (h < t->htsize);

      if (t->ind[h] == IDX_MAX)
	{
	  continue;
	}
      else if (t->ind[h] == ind[i])
	{
	  sim += t->val[h] * val[i];
	}
      else
	{			/* value may still be in overflow; use binary search to find it */
	  f = 0;
	  ASSERT (t->sz < t->maxsz);
	  l = t->sz - 1;
	  m = (f + l) / 2;
	  id = ind[i];
	  while (f <= l)
	    {
	      ASSERT (m < t->maxsz - 1);
	      if (t->data[m].key < id)
		{
		  f = m + 1;
		}
	      else if (t->data[m].key == id)
		{
		  sim += t->data[m].val * val[i];
		  break;
		}
	      else
		{
		  l = m - 1;
		}

	      m = (f + l) / 2;
	    }
	}
    }

  return sim;
}

inline val_t
sdhashVal (sdhash const *const t, idx_t const key)
{
  size_t register i;
  size_t register h;
  ssize_t register f;
  ssize_t register l;
  ssize_t register m;

  h = sdhashHash (key, t->htmask);
  ASSERT (h < t->htsize);
  if (t->ind[h] == IDX_MAX)
    {
      return 0.0;
    }
  else if (t->ind[h] == key)
    {
      return t->val[h];
    }
  else
    {				/* value may still be in overflow; use binary search to find it */
      f = 0;
      ASSERT (t->sz < t->maxsz);
      l = t->sz - 1;
      m = (f + l) / 2;
      while (f <= l)
	{
	  ASSERT (m < t->maxsz - 1);
	  if (t->data[m].key < key)
	    {
	      f = m + 1;
	    }
	  else if (t->data[m].key == key)
	    {
	      return t->data[m].val;
	    }
	  else
	    {
	      l = m - 1;
	    }

	  m = (f + l) / 2;
	}
    }

  return 0.0;
}



inline bool
sdhashIsset (sdhash const *const t, idx_t const key)
{
  size_t register i;
  size_t register h;
  ssize_t register f;
  ssize_t register l;
  ssize_t register m;

  h = sdhashHash (key, t->htmask);
  ASSERT (h < t->htsize);
  if (t->ind[h] == IDX_MAX)
    {
      return false;
    }
  else if (t->ind[h] == key)
    {
      return true;
    }
  else
    {				/* value may still be in overflow; use binary search to find it */
      f = 0;
      ASSERT (t->sz < t->maxsz);
      l = t->sz - 1;
      m = (f + l) / 2;
      while (f <= l)
	{
	  ASSERT (m < t->maxsz - 1);
	  if (t->data[m].key < key)
	    {
	      f = m + 1;
	    }
	  else if (t->data[m].key == key)
	    {
	      return true;
	    }
	  else
	    {
	      l = m - 1;
	    }

	  m = (f + l) / 2;
	}
    }

  return false;
}
