

#include "includes.h"




/*-------------------------------------------------------------
 * Timing inline functions
 *-------------------------------------------------------------*/
/**
 * Returns time stored in wall timer in seconds
 */
inline double
timer_get (double tmr)
{
  if (tmr < MAX_RUNTIME)
    {			      /** stop timer if started */
      tmr += da_wallclock ();
    }
  return tmr * 0.000001;
}

/**
 * Returns time stored in cpu timer in seconds
 */
inline double
cputimer_get (double tmr)
{
  if (tmr < MAX_RUNTIME)
    {			      /** stop timer if started */
      tmr += da_cpuclock ();
    }
  return tmr * 0.000001;
}

/*************************************************************************
* This function returns the CPU seconds
**************************************************************************/
double
da_wallclock (void)
{
#ifdef __GNUC__
  struct timeval ctime;

  gettimeofday (&ctime, NULL);

  return (1.0e+6 * (double) ctime.tv_sec + (double) ctime.tv_usec);
#else
  return (double) time (NULL);
#endif
}


/*************************************************************************
 * This function returns the CPU seconds
 **************************************************************************/
double
da_cpuclock (void)
{
#if defined(WIN32) || defined(__MINGW32__)
  return ((double) clock () / CLOCKS_PER_SEC);
#else
  struct rusage r;

  getrusage (RUSAGE_SELF, &r);
  return (1.0e+6 * (r.ru_utime.tv_sec + r.ru_stime.tv_sec) +
	  (r.ru_utime.tv_usec + r.ru_stime.tv_usec));
#endif
}


/*************************************************************************/
/*! Prints a timer in human readable format. Timers should be started and
 * stopped with da_startwctimer(t) and da_stopwctimer(t) respectively.
 *
 *  \param name is the timer name
 *  \param time is the time in seconds
 */
/*************************************************************************/
void
da_printTime (char const *const name, double time)
{
  time = timer_get (time);
  if (time < 60)
    {
      printf ("%s %.2fs\n", name, time);
    }
  else if (time < 3600)
    {
      int32_t min = time / 60;
      time -= min * 60;
      printf ("%s %dm %.2fs\n", name, min, time);
    }
  else
    {
      int32_t hours = time / 3600;
      time -= hours * 3600;
      int32_t min = time / 60;
      time -= min * 60;
      printf ("%s %dh %dm %.2fs\n", name, hours, min, time);
    }
}

void
da_printTimer (char const *const name, double time)
{
  da_printTime (name, timer_get (time));
}

/************************************************************************/
/*! \brief Prints a timer in long format, including human readable and
 * computer readable time. Timers should be started and stopped with
 * da_startwctimer(t) and da_stopwctimer(t) respectively.

    \param name is the Timer name, a string to be printed before the timer.
    \param time is the time that should be reported.
*/
/*************************************************************************/
void
da_printTimeLong (char const *const name, double time)
{
  double t = time;
  if (time < 60)
    {
      printf ("%s %.4f (%.2fs)\n", name, t, time);
    }
  else if (time < 3600)
    {
      int32_t min = time / 60;
      time -= min * 60;
      printf ("%s %.4f (%dm %.2fs)\n", name, t, min, time);
    }
  else
    {
      int32_t hours = time / 3600;
      time -= hours * 3600;
      int32_t min = time / 60;
      time -= min * 60;
      printf ("%s %.4f (%dh %dm %.2fs)\n", name, t, hours, min, time);
    }
}

void
da_printTimerLong (char const *const name, double time)
{
  da_printTimeLong (name, timer_get (time));
}


void
da_printThreadTimers (char const *const name,
		      idx_t const timerid, da_tsearch_t ** search)
{
  size_t i, n;
  double t, min, max, mean, li1, li2;

  n = search[0]->nthreads;
  t = max = mean = li1 = li2 = 0.0;
  min = DBL_MAX;
  printf ("\n%s:\n", name);

  for (i = 0; i < n; ++i)
    {

      switch (timerid)
	{
	case 1:
	  t = timer_get (search[i]->timer_1);
	  break;
	case 2:
	  t = timer_get (search[i]->timer_2);
	  break;
	case 3:
	  t = timer_get (search[i]->timer_3);
	  break;
	case 4:
	  t = timer_get (search[i]->timer_4);
	  break;
	case 5:
	  t = timer_get (search[i]->timer_5);
	  break;
	case 6:
	  t = timer_get (search[i]->timer_6);
	  break;
	case 7:
	  t = timer_get (search[i]->timer_7);
	  break;
	case 8:
	  t = timer_get (search[i]->timer_8);
	  break;
	case 9:
	  t = timer_get (search[i]->timer_9);
	  break;
	default:
	  da_errexit ("da_printThreadTimers: Wrong timer id.");
	  break;
	}

      printf ("%.4f  ", t);
      if (t < min)
	{
	  min = t;
	}
      if (t > max)
	{
	  max = t;
	}
      mean += t;

    }
  mean /= n;

  li1 = 100 * (max - mean) * n / ((n - 1) * max);
  li2 = 100 * (max / mean - 1.0);

  printf ("\nli1   %.4f\n", li1);
  printf ("li2   %.4f\n", li2);
  da_printTimeLong ("min  ", min);
  da_printTimeLong ("max  ", max);
  da_printTimeLong ("mean ", mean);

  fflush (stdout);
}
